from simframe.integration.scheme import Scheme

# Butcher coefficients
a10, a21 = 1/3, 2/3
c1, c2 = 1/3, 2/3
b0, b2 = 1/4, 3/4


def _f_expl_3_heun(x0, Y0, dx, *args, dYdx=None, **kwargs):
    """Explicit 3rd-order Heun's method

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    dYdx : Field, optional, default : None
        Current derivative. Will be calculated, if not set.
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated

    Butcher tableau
    ---------------
      0  |  0   0   0
     1/3 | 1/3  0   0
     2/3 |  0  2/3  0
    -----|-------------
         | 1/4  0  3/4
    """
    k0 = Y0.derivative(x0, Y0) if dYdx is None else dYdx
    k1 = Y0.derivative(x0 + c1*dx, Y0 + a10*k0*dx)
    k2 = Y0.derivative(x0 + c2*dx, Y0 + a21*k1*dx)

    return dx*(b0*k0 + b2*k2)


expl_3_heun = Scheme(
    _f_expl_3_heun, description="Explicit 3rd-order Heun's method")



from simframe.frame.updater import Updater


class Heartbeat(object):
    """This class controls an update including ``systole`` and ``diastole``.

    A full cardiac cycle constists of a systole operation, the actual update and a systole operation.
    All three are of type ``Updater``.

    The ``beat`` function calls ``systole``, ``updater``, ``diastole`` in that order and returns the return value
    of the ``updater``. Any positional or keyword arguments are only passed to the ``updater``, NOT to
    ``systole`` and ``diastole``.

    """

    __name__ = "Heartbeat"

    _systole = None
    _updater = None
    _diastole = None

    def __init__(self, updater=None, systole=None, diastole=None):
        """``Heartbeat`` class

        Parameters
        ----------
        updater : Updater, callable, None, optional, default : None
            This it the main updater of the heartbeat object
        systole : Updater, callable, None, optional, default : None
            The updater of the systole
        diastole : Updater, callable, None, optional, default : None
            The updater of the diastole"""

        self.updater = updater
        self.systole = systole
        self.diastole = diastole

    def __str__(self):
        return "{}".format(str(self.__name__))

    def __repr__(self):
        return self.__str__()

    @property
    def systole(self):
        '''``Updater`` that is called in the beginning of cardiac cycle.'''
        return self._systole

    @systole.setter
    def systole(self, value):
        if isinstance(value, Updater):
            self._systole = value
        elif hasattr(value, "__call__") or value is None:
            self._systole = Updater(value)
        else:
            raise TypeError(
                "Systole has to be of type Updater, None, or has to be callable.")

    @property
    def updater(self):
        '''``Updater`` that is performing the update instruction.'''
        return self._updater

    @updater.setter
    def updater(self, value):
        if isinstance(value, Updater):
            self._updater = value
        elif hasattr(value, "__call__") or value is None:
            self._updater = Updater(value)
        else:
            raise TypeError(
                "Updater has to be of type Updater, None, or has to be callable.")

    @property
    def diastole(self):
        '''``Updater`` that is called at the end of the cardiac cycle.'''
        return self._diastole

    @diastole.setter
    def diastole(self, value):
        if isinstance(value, Updater):
            self._diastole = value
        elif hasattr(value, "__call__") or value is None:
            self._diastole = Updater(value)
        else:
            raise TypeError(
                "Diastole has to be of type Updater, None, or has to be callable.")

    def beat(self, owner, *args, Y=None, **kwargs):
        """This method executes ``systole``, ``updater``, and ``distole`` in that order and returns the return value of
        the ``updater``.

        Parameters
        ----------
        owner : Frame
            Parent frame object to which updater belongs
        Y : Field, optional, default : None
            Field that should be updated

        Notes
        -----
            *args and **kwargs are only passed to updater, NOT to systole and diastole

        Returns
        -------
        ret : Return value of updater."""

        # Perform systole operation
        self.systole.update(owner)

        # Perform update operation and save return value.
        ret = self.updater.update(owner, *args, **kwargs)
        # If Y is given and a return value was received, then beat is updating a field.
        # The new field value needs to be set here, so the diastole has access to the new value.
        if Y is not None and ret is not None:
            Y[...] = ret

        # Perform diastole operation.
        self.diastole.update(owner)
        return ret



from simframe.integration.scheme import Scheme

import numpy as np
from scipy.sparse import linalg


def _f_impl_1_euler_gmres(x0, Y0, dx, jac=None, gmres_opt={}, *args, **kwargs):
    """Implicit 1st-order Euler integration scheme with GMRES solver

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    jac : Field, optional, defaul : None
        Current Jacobian. Will be calculated, if not set
    gmres_opt : dict, optional, default : {}
        dictionary with options for scipy GMRES solver
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated

    Butcher tableau
    ---------------
     1 | 1
    ---|---
       | 1 
    """
    if jac is None:
        jac = Y0.jacobian(x0 + dx)
    N = jac.shape[0]
    eye = np.eye(N)
    A = eye - dx*jac
    res, state = linalg.gmres(A, Y0, **gmres_opt)
    if state != 0:
        return False
    else:
        return res - Y0


impl_1_euler_gmres = Scheme(
    _f_impl_1_euler_gmres, description="Implicit 1st-order Euler method with GMRES solver")



import dill
import numpy as np
import os

from simframe.frame.field import Field
from simframe.frame.intvar import IntVar
from simframe.io.reader import Reader
from simframe.io.dump import writedump
from simframe.frame.abstractgroup import AbstractGroup
from simframe.utils.color import colorize


class Writer(object):
    """General class for writing output files. It should be used as wrapper for customized ``Writer``."""

    __name__ = "Writer"

    _datadir = "data"
    _description = ""
    _dumping = True
    _extension = "out"
    _filename = "data"
    _options = {}
    _read = None
    _overwrite = False
    _verbosity = 1
    _zfill = 4

    def __init__(self, func, datadir="data", filename="data", zfill=4, extension="out", overwrite=False, dumping=True,
                 reader=None, verbosity=1, description="", options={}):
        """Parameters
        ----------
        func : callable
            Function that is writing the outputs to file
        datadir : string, optional, default : "data"
            relative path to data directory
        filename : string, optional, default : "data"
            basic name of the files to be written
        zfill : int, optional, default : 4
            leading zeros of numbered files
        extension : string, optional, default : "out"
            filename extension of the data files
        overwrite : boolean, optional, default : False
            If existing files should be overwritten
        dumping : boolean, optional, default : True
            If True dump files will be written
        reader : Reader, optional, default : None
            reader to read files
        verbosity : int, optional, default : 1
            Verbosity of writer
        options : dict, optional, default : {}
            Optional keyword arguments that need to be passed to writing algorithm"""
        self._func = func
        self.datadir = datadir
        self.filename = filename
        self.zfill = zfill
        self.extension = extension
        self.overwrite = overwrite
        self.dumping = dumping
        self.description = description
        self.options = options
        self.verbosity = verbosity
        self.read = reader(self) if reader is not None else None

    @property
    def datadir(self):
        '''Data directory of output files.'''
        return self._datadir

    @datadir.setter
    def datadir(self, value):
        if not isinstance(value, str):
            raise TypeError("datadir has to be of type str.")
        self._datadir = value

    @property
    def description(self):
        '''Description of ``Writer``.'''
        return self._description

    @description.setter
    def description(self, value):
        if value is None:
            self._description = ""
        if not isinstance(value, str):
            raise TypeError("description has to be of type str or None.")
        self._description = value

    @property
    def extension(self):
        '''Filename extension of output files.'''
        return self._extension

    @extension.setter
    def extension(self, value):
        if not isinstance(value, str):
            raise TypeError("extension has to be of type str.")
        if value.startswith("."):
            raise ValueError("extension should not start with '.'")
        self._extension = value

    @property
    def filename(self):
        '''Base filename of output files.'''
        return self._filename

    @filename.setter
    def filename(self, value):
        if not isinstance(value, str):
            raise TypeError("filename has to be of type str.")
        if value == "":
            raise ValueError("filename cannot be empty.")
        self._filename = value

    @property
    def options(self):
        '''Dictionary of keyword arguments passed to customized writing routine.'''
        return self._options

    @options.setter
    def options(self, value):
        if not isinstance(value, dict):
            raise TypeError("options has to be of type dict.")
        self._options = value

    @property
    def read(self):
        '''``Reader`` object for reading output files.'''
        return self._reader

    @read.setter
    def read(self, value):
        if isinstance(value, Reader) or (value is None):
            self._reader = value
        else:
            raise TypeError("<reader> has to be of type Reader or None.")

    @property
    def overwrite(self):
        '''If ``True`` existing output files will be overwritten.'''
        return self._overwrite

    @overwrite.setter
    def overwrite(self, value):
        if not isinstance(value, np.int):
            raise TypeError("<overwrite> has to be of type bool.")
        if value:
            self._overwrite = True
        else:
            self._overwrite = False

    @property
    def dumping(self):
        '''If ``True`` dump files will be written.'''
        return self._dumping

    @dumping.setter
    def dumping(self, value):
        if not isinstance(value, np.int):
            raise TypeError("<dumping> has to be of type bool.")
        if value:
            self._dumping = True
            self._preparedump()
        else:
            self._dumping = False

    @property
    def verbosity(self):
        '''Verbosity of the writer.'''
        return self._verbosity

    @verbosity.setter
    def verbosity(self, value):
        if not isinstance(value, np.int):
            raise TypeError("<verbosity> has to be of type int.")
        self._verbosity = value

    @property
    def zfill(self):
        '''Zero padding of numbered files names.'''
        return self._zfill

    @zfill.setter
    def zfill(self, value):
        value = np.int(value)
        if not isinstance(value, np.int):
            raise TypeError("zfill has to be integer.")
        self._zfill = value

    def checkdatadir(self, datadir=None, createdir=False):
        """Function checks if data directory exists and creates it if necessary.

        Parameters
        ----------
        datadir : string or None, optinal, default : None
            Data directory to be checked. If None it assumes the data directory of the parent writer.
        createdir : boolen, optional, default : False
            If True function creates data directory if it does not exist.

        Returns
        -------
        datadirexists : boolean
            True if directory exists, False if not"""

        datadir = self.datadir if datadir is None else datadir

        if not os.path.exists(self.datadir) and createdir:
            if self.verbosity > 0:
                msg = "Creating data directory '{:s}'.".format(self.datadir)
                print(msg)
            os.makedirs(self.datadir)

        return os.path.exists(self.datadir)

    def __str__(self):
        ret = AbstractGroup.__str__(self)
        return ret

    def __repr__(self):
        ret = self.__str__()+"\n"
        ret += "-" * (len(ret)-1) + "\n"
        ret += "    Data directory : {}\n".format(self.datadir)
        ret += "    File names     : {}\n".format(self._getfilename(0))
        ret += "    Overwrite      : {}\n".format(
            colorize(self.overwrite, "yellow") if self.overwrite else self.overwrite)
        ret += "    Dumping        : {}\n".format(
            colorize(self.dumping, "yellow") if not self.dumping else self.dumping)
        ret += "    Options        : {}\n".format(self.options)
        ret += "    Verbosity      : {}".format(self.verbosity)
        return ret

    def _getfilename(self, i):
        """This function creates ``<path>/<filename>`` for a given output number

        Parameters
        ----------
        i : integer
            Number of output

        Returns
        -------
        filename : str
            The constructed filename"""

        # Removing . from extension if given
        ext = self.extension
        if ext != "":
            if ext[0] != ".":
                ext = "." + ext

        number = str(i).zfill(self.zfill)
        filename = self.filename + number + ext

        return os.path.join(self.datadir, filename)

    def writedump(self, frame, filename=""):
        """Writes the ``Frame`` to dump file

        Parameters
        ----------
        frame : object
            object to be written to file
        filename : str, optional, default : ""
            path to file to be written
            if not set, filename will be <writer.datadir>/frame.dmp."""

        filename = os.path.join(
            self.datadir, "frame.dmp") if filename == "" else filename
        self.checkdatadir(createdir=True)

        if self.verbosity > 0:
            msg = "Writing dump file {}".format(colorize(filename, "blue"))
            print(msg)

        writedump(frame, filename)

    def _preparedump(self):
        """This function dumps a template ``<simframe.frame.Field>`` and ``<simframe.frame.IntVar>`` to file
        and deletes the file directly afterwards. For some reason ``dill`` does not correctly save ``Field`` and ``IntVar``
        correctly the first time. This function is called everytime ``<Writer.dumping>`` is set to ``True``
        to make sure dumping works correctly when needed.

        This is a "dirty fix". The reason why ``dill`` does not work correctly the first time is unknown.
        """
        filename = "temp.dmp"
        temp = IntVar(None, 0., snapshots=[0.])
        writedump(temp, filename)
        temp = Field(None, 0.)
        writedump(temp, filename)
        os.remove(filename)

    def write(self, owner, i, forceoverwrite, filename=""):
        """Writes output to file

        Parameters
        ----------
        owner : Frame
            Parent ``Frame`` object
        i : int
            Number of output
        forceoverwrite : boolean
            If ``True`` it will forces and overwrite of the file if it exists.
        filename : string
            If this is not "" the writer will use this filename instead of the standard scheme"""

        if filename == "":
            filename = self._getfilename(i)
            self.checkdatadir(createdir=True)
        if not forceoverwrite:
            if not self.overwrite:
                if os.path.isfile(filename):
                    raise RuntimeError(
                        "File {} already exists.".format(filename))
        self._func(owner, filename, **self.options)
        if self.verbosity > 0:
            msg = "Writing file {}".format(colorize(filename, "blue"))
            print(msg)
        if self.dumping:
            self.writedump(owner)



from simframe.integration.scheme import Scheme

# Butcher coefficients
a10 = 0.4
a20, a21 = 0.29697761,  0.15875964
a30, a31, a32 = 0.21810040, -3.05096516, 3.83286476
b0, b1, b2, b3 = 0.17476028, -0.55148066, 1.20553560, 0.17118478
c1, c2 = 0.4,  0.45573725


def _f_expl_4_ralston(x0, Y0, dx, *args, dYdx=None, **kwargs):
    """Explicit 4th-order Ralston's method

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    dYdx : Field, optional, default : None
        Current derivative. Will be calculated, if not set.
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated

    Butcher tableau
    ---------------
     0          | 0           0          0          0
     0.4        | 0.4         0          0          0
     0.45573725 | 0.29697761  0.15875964 0          0
     1          | 0.21810040 -3.05096516 3.83286476 0
    ------------|----------------------------------------------
                | 0.17476028 -0.55148066 1.20553560 0.17118478
    """

    k0 = Y0.derivative(x0, Y0) if dYdx is None else dYdx
    k1 = Y0.derivative(x0 + c1*dx, Y0 + a10*k0 * dx)
    k2 = Y0.derivative(x0 + c2*dx, Y0 + (a20*k0 + a21*k1)*dx)
    k3 = Y0.derivative(x0 + dx, Y0 + (a30*k0 + a31*k1 + a32*k2)*dx)

    return dx*(b0*k0 + b1*k1 + b2*k2 + b3*k3)


expl_4_ralston = Scheme(
    _f_expl_4_ralston, description="Explicit 4th-order Ralston's method")



from collections import deque
import numpy as np

from simframe.frame.abstractgroup import AbstractGroup
from simframe.frame.heartbeat import Heartbeat
from simframe.frame.intvar import IntVar
from simframe.integration.instruction import Instruction
from simframe.integration.schemes import update


class Integrator:
    """``Integrator`` class that manages the integration instructions"""

    __name__ = "Integrator"

    _description = ""
    _instructions = []
    _failop = Heartbeat(None)
    _finalizer = Heartbeat(None)
    _maxit = 500
    _preparator = Heartbeat(None)
    _var = None

    def __init__(self, var, instructions=[], failop=None, preparator=None, finalizer=None, maxit=500, description=""):
        """Integrator

        Parameters
        ----------
        var : IntVar
            Integration variable that controls the integration
        instructions : list of Instruction, optional, default : None
            List of the instegration instructions
        failop : Heartbeat, Updater, callable, or None, optional, default : None
            Fail operation that is executed if any instruction failed
        preparator : Heartbeat, Updater, callable, or None, optional, default : None
            Heartbeat that will be executed before the integration
        finalizer : Heartbeat, Updater, callable, or None, optional, default : None
            Heartbeat that will be executed after the integration
        maxit : int, optional, default : 5000
            Maximum number of integration iterations
        description : str, optional, default : ""
            Description of integrator"""
        self.description = description
        self.failop = failop
        self.finalizer = finalizer
        self.instructions = instructions
        self.maxit = maxit
        self.preparator = preparator
        self.var = var

    def __str__(self):
        return AbstractGroup.__str__(self)

    def __repr__(self):
        return self.__str__()

    @property
    def description(self):
        '''Description of integrator'''
        return self._description

    @description.setter
    def description(self, value):
        if not isinstance(value, str):
            raise ValueError("<value> has to be of type str.")
        self._description = value

    @property
    def failop(self):
        '''``Heartbeat`` objects that is called if any integration ``Instruction`` returned ``False``'''
        return self._failop

    @failop.setter
    def failop(self, value):
        if isinstance(value, Heartbeat):
            self._failop = value
        else:
            self._failop = Heartbeat(value)

    @property
    def finalizer(self):
        '''``Heartbeat`` object that is called after the integration was successful.'''
        return self._finalizer

    @finalizer.setter
    def finalizer(self, value):
        if isinstance(value, Heartbeat):
            self._finalizer = value
        else:
            self._finalizer = Heartbeat(value)

    @property
    def instructions(self):
        '''List of integration ``Instructions`` that will be executed in that order.'''
        return self._instructions

    @instructions.setter
    def instructions(self, value):
        if not isinstance(value, list):
            raise TypeError("<instructions> has to be of type list.")
        for val in value:
            if not isinstance(val, Instruction):
                raise TypeError(
                    "<instructions> has to be list of Instructions")
        self._instructions = value

    @property
    def maxit(self):
        '''Maximum number of integration tries until program will be aborted.'''
        return self._maxit

    @maxit.setter
    def maxit(self, value):
        value = np.int(value)
        if not isinstance(value, np.int):
            raise TypeError("maxit has to be of type int.")
        if value <= 0:
            raise ValueError("maxit has to be larger 0.")
        self._maxit = value

    @property
    def preparator(self):
        '''``Heartbeat`` object that is called before the integration instructions will be executed.'''
        return self._preparator

    @preparator.setter
    def preparator(self, value):
        if isinstance(value, Heartbeat):
            self._preparator = value
        else:
            self._preparator = Heartbeat(value)

    @property
    def var(self):
        '''The integration variable ``IntVar`` that is associated with this ``Integrator``.'''
        return self._var

    @var.setter
    def var(self, value):
        if not isinstance(value, IntVar):
            raise TypeError("<var> has to be of type Intvar.")
        self._var = value

    def integrate(self):
        """Method that executes one integration step."""
        # Preparation
        self._prepare()
        # Loop over all instructions. Exit the loop only if all instructions were executed successfully
        # And count the loops
        i = 0
        status = False
        while(not status):
            # The suggested stepsize has to be reset in the beginning of every try.
            # We therefore have to copy the current stepsize in case the user is returning
            # the suggested stepsize in the stepsize function.
            stepsize = self.var.stepsize.copy()
            self.var._suggested = None
            if i >= self.maxit:
                raise StopIteration(
                    "Maximum number of integration attempts exceeded.")
            # Safe all return values in list
            ret = deque([])
            for inst in self.instructions:
                ret.append(inst(stepsize))
            # If no instruction returned False, Integration was successful. Exit the loop.
            if not np.any(np.array(ret) == False):
                status = True
            else:
                # Reset buffers if integration failed
                for inst in self.instructions:
                    inst.Y._buffer = None
                self._failoperation()
                i += 1
        # Update the variables.
        for i, inst in enumerate(self.instructions):
            if inst.Y._buffer is None:
                continue
            update(None, inst.Y, None)
        # Store the taken stepsize
        self.var._prevstepsize = stepsize
        # Finalization
        self._finalize()

    def _failoperation(self, *args, **kwargs):
        """This operation will be executed if any integration ``Instruction`` failed and before the
        ``Integrator`` tries it again. It will execute the ``Heartbeat`` of ``Integrator.failop``.

        Parameters
        ----------
        args : additional positional arguments
        kwargs : additional keyword arguments

        Notes
        -----
        args, and kwargs will only be passed to the ``updater``, NOT ``systole`` and ``diastole``."""
        self.failop.beat(self.var._owner, *args, **kwargs)

    def _prepare(self, *args, **kwargs):
        """This operation will be executed before the integration. It will execute the
        ``Heartbeat`` of ``Integrator.preparator``.

        Parameters
        ----------
        args : additional positional arguments
        kwargs : additional keyword arguments

        Notes
        -----
        args, and kwargs will only be passed to the ``updater``, NOT ``systole`` and ``diastole``."""
        self.preparator.beat(self.var._owner, *args, **kwargs)

    def _finalize(self, *args, **kwargs):
        """This operation will be executed before the integration. It will execute the
        ``Heartbeat`` of ``Integrator.finalizer``.

        Parameters
        ----------
        args : additional positional arguments
        kwargs : additional keyword arguments

        Notes
        -----
        args, and kwargs will only be passed to the ``updater``, NOT ``systole`` and ``diastole``."""
        self.finalizer.beat(self.var._owner, *args, **kwargs)



from simframe.integration.scheme import Scheme

# Butcher coefficients
a10 = 1/2
a21 = 3/4
b0, b1, b2 = 2/9, 1/3, 4/9
c1, c2 = 1/2, 3/4


def _f_expl_3_ralston(x0, Y0, dx, *args, dYdx=None, **kwargs):
    """Explicit 3rd-order Ralston's method

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    dYdx : Field, optional, default : None
        Current derivative. Will be calculated, if not set.
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated

    Butcher tableau
    ---------------
      0  |  0   0   0
     1/2 | 1/2  0   0
     3/4 |  0  3/4  0
    -----|-------------
         | 2/9 1/3 4/9
    """
    k0 = Y0.derivative(x0, Y0) if dYdx is None else dYdx
    k1 = Y0.derivative(x0 + c1*dx, Y0 + a10*k0*dx)
    k2 = Y0.derivative(x0 + c2*dx, Y0 + a21*k1*dx)

    return dx*(b0*k0 + b1*k1 + b2*k2)


expl_3_ralston = Scheme(
    _f_expl_3_ralston, description="Explicit 3rd-order Ralston's method")



import numpy as np
from simframe.frame.abstractgroup import AbstractGroup
from simframe.frame.heartbeat import Heartbeat
from simframe.utils.color import colorize


class Field(np.ndarray, AbstractGroup):
    """Class for storing simulation quantities.

    In addition to ``Group``, ``Field`` can have an ``differentiator`` for calculating its derivative and/or an
    ``jacobinator`` for calculating its Jacobian. The function that is calculating the derivative needs the parent
    ``Frame`` object as first, the integration variable of type ``IntVar`` as second, and the ``Field`` itself as
    third positional argument

    ``Field`` behaves like ``numpy.ndarray`` and can perform the same numerical operations.

    Notes
    -----
    When ``Field.update()`` is called ``Field`` will be updated according return value of the ``updater`` of the
    ``Heartbeat`` object assigned to the ``Field``. The function that is updating ``Field`` needs the parent ``Frame``
    object as first positional argument."""

    __name__ = "Field"

    _differentiator = Heartbeat(None)
    _jacobinator = Heartbeat(None)
    _constant = False
    _save = True
    _buffer = None

    def __new__(cls, owner, value, updater=None, differentiator=None, jacobinator=None, description="", constant=False, save=True, copy=False):
        """Parameters
        ----------
        owner : Frame
            Parent frame object to which the field belongs
        value : number, array, string
            Initial value of the field. Needs to have correct type and shape
        updater : Heartbeat, Updater, callable or None, optional, default : None
            Instruction for field update
        differentiator : Heartbeat, Updater, callable or None, optional, default : None
            Instruction for calculating derivative
        jacobinator : Heartbeat, Updater, callable or None, optional, default : None
            Instruction for calculating the Jacobi matrix
        description : string, optional, default : ""
            Descriptive string for the field
        constant : boolean, optional, default : False
            True if the field is immutable.
        save : boolean, optional, default : True
            If False the field is not written into output files
        copy : boolean, optional, default : False
            If True <value> will be copied, not referenced"""
        obj = np.array(value, copy=copy).view(cls)
        if obj.shape == ():
            obj = np.array([value], copy=copy).view(cls)
        obj._owner = owner
        obj.updater = Heartbeat(updater)
        obj.differentiator = Heartbeat(differentiator)
        obj.jacobinator = Heartbeat(jacobinator)
        obj.description = description
        obj.constant = constant
        obj.save = save
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self._owner = getattr(obj, "_owner", None)
        self.updater = getattr(obj, "updater", Heartbeat(None))
        self.differentiator = getattr(obj, "differentiator", Heartbeat(None))
        self.jacobinator = getattr(obj, "jacobinator", Heartbeat(None))
        self.description = getattr(obj, "description", "")
        self.constant = getattr(obj, "constant", False)

    def __str__(self):
        ret = AbstractGroup.__str__(self)
        if self.constant:
            ret += ", {}".format(colorize("constant", "purple"))
        return ret

    def __repr__(self):
        val = self.getfield(self.dtype)
        if val.shape == (1,):
            val = np.array(val[0])
        ret = "{}".format(np.ndarray.__str__(val))
        return ret

    def __format__(self, format_spec):
        if self.shape == (1,):
            return self[0].__format__(format_spec)
        else:
            return super().__format__(format_spec)

    @property
    def constant(self):
        '''If True, ``Field`` is immutable.'''
        return self._constant

    @constant.setter
    def constant(self, value):
        if isinstance(value, np.int):
            if value:
                self._constant = True
            else:
                self._constant = False
        else:
            raise TypeError("<value> hat to be of type bool.")

    @property
    def save(self):
        '''If False, ``Field`` will not be stored in output files.'''
        return self._save

    @save.setter
    def save(self, value):
        if isinstance(value, np.int):
            if value:
                self._save = True
            else:
                self._save = False
        else:
            raise TypeError("<value> hat to be of type bool.")

    @property
    def buffer(self):
        '''Temporary buffer that stores the new value of ``Field`` after successful integration.'''
        return self._buffer

    @buffer.setter
    def buffer(self, value):
        raise RuntimeError("Do not set buffer directly.")

    @property
    def differentiator(self):
        '''``Heartbeat`` object with instructions for calculating the derivative of ``Field``'''
        return self._differentiator

    @differentiator.setter
    def differentiator(self, value):
        if isinstance(value, Heartbeat):
            self._differentiator = value
        else:
            self._differentiator = Heartbeat(value)

    @property
    def jacobinator(self):
        '''``Heartbeat`` object with instructions for calculating the Jacobian of ``Field``'''
        return self._jacobinator

    @jacobinator.setter
    def jacobinator(self, value):
        if isinstance(value, Heartbeat):
            self._jacobinator = value
        else:
            self._jacobinator = Heartbeat(value)

    def update(self, *args, **kwargs):
        """Function to update the ``Field``.

        Parameter
        ---------
        args : additional positional arguments
        kwargs : additional keyword arguments

        Notes
        -----
        Function calls the Heartbeat object of the ``Field``. Additional positional and keyword arguments are only
        passed to the ``updater``, NOT to ``systole`` and ``diastole``."""
        self.updater.beat(self._owner, *args, Y=self, **kwargs)

    def derivative(self, x=None, Y=None, *args, **kwargs):
        """If ``differentiator`` or ``jacobinator`` is set, this returns the derivative of the ``Field``.

        Parameters
        ----------
        x : IntVar, optional, default : None
            Integration variable
            If None it uses the integration variable of the integrator of the parent Frame
        Y : Field, optional, default : None
            Derivative of Y with respect to the integration variable.
            If None it uses the field itself

        Returns
        -------
        deriv : derivative of the field according the differetiator or jacobinator

        Notes
        -----
        The function that calculates the derivative needs the parent ``Frame`` as first positional, the
        integration variable ``IntVar`` as second positional, and the ``Field`` itself as third positional argument.

        The ``differentiator`` is not set, it will try to calculate the derivative from the Jacobian.
        If ``jacobinator`` is also not set, it will return ``False``"""
        if x is None:
            if self._owner.integrator is None:
                raise RuntimeError("x not given and no integrator set.")
            if self._owner.integrator.var is None:
                raise RuntimeError(
                    "x not given and no integration variable set in integrator.")
            x = self._owner.integrator.var
        Y = Y if Y is not None else self
        deriv = self.differentiator.beat(self._owner, x, Y, *args, **kwargs)
        if deriv is not None:
            return deriv
        jac = self.jacobinator.beat(self._owner, x)
        if jac is not None:
            return jac @ Y
        else:
            # If no differentiator or jacobian is set we return zeros.
            return np.zeros_like(self)

    def jacobian(self, x=None, *args, **kwargs):
        """If ``jacobinator`` is set, this returns the Jacobi matrix of the ``Field``.

        Parameters
        ----------
        x : IntVar, optional, default : None
            Integration variable
            If None it uses the integration variable of the integrator of the parent Frame

        Returns
        -------
        jac : Jacobi matrix of the field according the differetiator

        The function that calculates the Jacobian needs the parent frame as first positional and the
        integration variable as second positional."""
        if x is None:
            if self._owner.integrator is None:
                raise RuntimeError("x not given and no integrator set.")
            if self._owner.integrator.var is None:
                raise RuntimeError(
                    "x not given and no integration variable set in integrator.")
            x = self._owner.integrator.var
        return self.jacobinator.beat(self._owner, x, *args, **kwargs)

    def _setvalue(self, value):
        """Function to set a value to the field. Direct assignement of values does overwrite the Field object.

        Parameters
        ----------
        value : number, array, string
            Value to be written into the field. Needs to have correct type and shape"""
        if self._constant:
            raise RuntimeError("Field is constant.")
        value = np.asarray(value)
        if value.shape == ():
            value = np.array([value])
        self.setfield(value, self.dtype)



from simframe import Frame
from simframe.frame import Field

import numpy as np
import pytest


def test_field_creationcopy():
    A = np.ones(5)
    f = Frame()
    f.addfield("f", A)
    assert np.all(f.f == A)
    A *= 2
    assert np.all(f.f == np.ones(5))
    f.addfield("f", A, copy=False)
    assert np.all(f.f == A)
    A *= 2
    assert np.all(f.f == A)


def test_field_assignment():
    A = np.ones(5)
    B = np.ones(5) * 2.
    f = Frame()
    f.addfield("f", A)
    assert np.all(f.f == A)
    f.f = B
    assert isinstance(f.f, Field)
    assert np.all(f.f == B)
    B *= 2.
    assert not np.all(f.f == B)
    with pytest.raises(ValueError):
        f.f = 1
    f.addfield("f", 0)
    with pytest.raises(ValueError):
        f.f = "test"


def test_field_update():
    f = Frame()
    f.addfield("f", np.ones(5))
    f.update()
    assert np.all(f.f == np.ones(5))

    def upd(f):
        return 2. * f.f
    f.f.updater = upd
    f.f.update()
    assert np.all(f.f == 2. * np.ones(5))


def test_field_differentiator():
    f = Frame()
    f.addfield("f", np.ones(5))

    def upd(f, x, Y):
        return 2. * x * Y
    f.f.differentiator = upd
    assert np.all(f.f.derivative(x=2.) == 4. * np.ones(5))



from simframe.integration.scheme import Scheme


def _f_expl_1_euler(x0, Y0, dx, *args, dYdx=None, **kwargs):
    """Explicit 1st-order Euler integration scheme

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    dYdx : Field, optional, default : None
        Current derivative. Will be calculated, if not set.
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated

    Butcher tableau
    ---------------
     0 | 0
    ---|---
       | 1 
    """
    k0 = Y0.derivative(x0, Y0) if dYdx is None else dYdx
    return dx*k0


expl_1_euler = Scheme(
    _f_expl_1_euler, description="Explicit 1st-order Euler method")



from functools import partial

from simframe.frame.abstractgroup import AbstractGroup
from simframe.frame.field import Field
from simframe.frame.intvar import IntVar
from simframe.frame.heartbeat import Heartbeat
from simframe.utils.color import colorize


class Group(AbstractGroup):
    """Class for grouping data. ``Group`` is a data frame that has additional functionality for updating its attributes.

    Notes
    -----
    When ``Group.update()`` is called the instructions of the group's ``Heartbeat`` object will be performed.
    The function that is determing the update operation needs the parent ``Frame`` object as first positional argument."""

    __name__ = "Group"

    def __init__(self, owner, updater=None, description=""):
        """Parameters
        ----------
        owner : Frame
            Parent frame object to which the group belongs
        updater : Heartbeat, Updater, callable, list or None, optional, default : None
            Updater for group update. A Heartbeat object will be created from this.
        description : string, optional, default : ""
            Descriptive string for the group

        Notes
        -----
        The updater of groups can take a list of string with the attribute names that should be updated
        in the order in which they should be updated. It will create a callable function from that list."""
        self._description = description
        self._owner = owner
        self.updater = updater

    def __setattr__(self, name, value):
        """Function to set an attribute including fields.
        This function allows the user to change the value of fields instead of replacing them."""
        if name in self.__dict__ and isinstance(self.__dict__[name], Field):
            self.__dict__[name][...] = value
        else:
            super().__setattr__(name, value)

    def __repr__(self):
        """Function to have good looking overview of the members of the group."""

        fields = {}
        groups = {}
        misc = {}

        # return value
        ret = ""

        for key, val in self.__dict__.items():
            # Don't show private attributes
            if key.startswith("_"):
                continue

            # Sort attributes by group, field and else
            if isinstance(val, Field):
                fields[key] = val
            elif isinstance(val, Group):
                groups[key] = val
            else:
                misc[key] = val

        # Underlined headline. The length of the underline is off if there are hidden characters, like color.
        ret += self.__str__() + "\n"
        ret += "-" * (len(ret) - 1) + "\n"

        # Printing all groups alphanumerically sorted by name
        if len(groups) > 0:
            for key in sorted(groups.keys(), key=str.casefold):
                if len(key) > 12:
                    name = key[:9] + "..."
                else:
                    name = key
                ret += "    {:12s} : {}\n".format(name, groups[key])
            ret += "  -----\n"

        # Printing all fields alphanumerically sorted by name
        if len(fields) > 0:
            for key in sorted(fields.keys(), key=str.casefold):
                if len(key) > 12:
                    name = key[:9] + "..."
                else:
                    name = key
                ret += "    {:12s} : {}\n".format(name, fields[key].__str__())
            ret += "  -----\n"

        # Printing everything else alphanumerically sorted
        if len(misc) > 0:
            for key in sorted(misc.keys(), key=str.casefold):
                if len(key) > 12:
                    name = key[:9] + "..."
                else:
                    name = key
                ret += "    {:12s} : {}\n".format(name,
                                                  type(misc[key]).__name__)
            ret += "  -----\n"

        # The Frame object should have an integrator and writer which are displayed separately.
        # If the object has an integrator
        if "_integrator" in self.__dict__.keys():
            integrator = self.__dict__["_integrator"]
            # If not set, print warning
            txt = colorize("not specified", "yellow")
            if integrator is not None:
                txt = integrator.__str__()
            ret += "    {:12s} : {}".format("Integrator", txt)
            ret += "\n"

        # If the object has a writer
        if "_writer" in self.__dict__.keys():
            writer = self.__dict__["_writer"]
            # If not set print warning
            txt = colorize("not specified", "yellow")
            if writer is not None:
                txt = writer.__str__()
            ret += "    {:12s} : {}".format("Writer", txt)
            ret += "\n"

        return ret

    # We need to overwrite the updater property of AbstractGroup, because we want the group to be able
    # to take lists of attributes as value.
    @property
    def updater(self):
        '''``Heartbeat`` object with update instructions.

        You can either set a ``Heartbeat`` object directly, a callable functions that will be automatically transformed into
        a ``Heartbeat`` object, or a list of attribute names of the ``Group`` that will be updated in that order.'''
        return self._updater

    @updater.setter
    def updater(self, value):
        if isinstance(value, Heartbeat):
            self._updater = value
        elif isinstance(value, list):
            self._checkupdatelist(value)
            self._updater = Heartbeat(self._createupdatefromlist(value))
        else:
            self._updater = Heartbeat(value)

    @property
    def toc(self):
        '''Complete table of contents starting from this object.'''
        self._toc()

    @toc.setter
    def toc(self, value):
        pass

    def addfield(self, name, value, updater=None, differentiator=None, description="", constant=False, copy=True):
        """Function to add a new ``Field`` to the object.

        Parameters
        ----------
        name : string
            Name of the field
        value : number, array, string
            Initial value of the field. Needs to have already the correct type and shape
        updater : Heartbeat, Updater, callable or None, optional, default : None
            Updater for field update
        differentiator : Heartbeat, Updater, callable or None, optional, default : None
            Differentiator if the field has a derivative
        description : string, optional, default : ""
            Descriptive string for the field
        constant : boolean, optional, default : False
            True if the field is immutable
        copy : boolean, optional, default : True
            If True <value> will be copied, not referenced
        """
        self.__dict__[name] = Field(self._owner, value, updater=updater,
                                    differentiator=differentiator, description=description, constant=constant, copy=copy)

    def addgroup(self, name, updater=None, description=""):
        """Function to add a new ``Group`` to the object.

        Parameters
        ----------
        name : string
            Name of the group
        updater : Heartbeat, Updater, callable or None, optional, default : None
            Updater for field update
        description : string, optional, default : ""
            Descriptive string for the group
        """
        self.__dict__[name] = Group(
            self._owner, updater=updater, description=description)

    def addintegrationvariable(self, name, value, snapshots=[], updater=None, description="", copy=True):
        """Function to add a new integration variable ``IntVar`` to the object.

        Parameters
        ----------
        name : string
            Name of the field
        value : number, array, string
            Initial value of the field. Needs to have already the correct type and shape
        updater : Heartbeat, Updater, callable or None, optional, default : None
            Updater for field update
        snapshots : list, ndarray, optional, default : []
            List of snapshots at which an output file should be written
        description : string, optional, default : ""
            Descriptive string for the field
        copy : boolean, optional, default : True
            If True <value> will be copied, not referenced
        """
        self.__dict__[name] = IntVar(
            self._owner, value, updater=updater, snapshots=snapshots, description=description, copy=copy)

    def _checkupdatelist(self, ls):
        """This function checks if a list is suitable to be used as update.

        Parameters
        ----------
        ls : list
            list of string with the attribute names that should be updated in that order"""
        for val in ls:
            if not isinstance(val, str):
                raise ValueError("List has to be list of strings.")
        for val in ls:
            if val not in self.__dict__:
                msg = "{}: {} is not an attribute of the group".format(colorize("Warning", "yellow"),
                                                                       val)
                print(msg)

    def _createupdatefromlist(self, ls):
        """This method creates an update method from a list.

        Parameters
        ----------
        ls : list
            list of group attributes that should be updated in that order

        Returns
        -------
        func : callable
            Function that is reduced by <self> and <ls>."""
        return partial(_dummyupdatewithlist, self, ls)

    def _toc(self):
        ret = _toc_tree(self)
        print(ret)


def _toc_tree(obj, prefix=""):
    ret = colorize(obj.__str__(), "blue")
    prefix = prefix + 4*" "
    for key in sorted(obj.__dict__.keys(), key=str.casefold):
        if key.startswith("_"):
            continue
        val = obj.__dict__[key]
        ret += "\n{}- {}: ".format(
            prefix, colorize(key, "blue"))
        if isinstance(val, Group):
            ret += _toc_tree(val, prefix=prefix)
        else:
            ret += val.__str__()
    return ret


def _dummyupdatewithlist(grp, ls, owner, *args, **kwargs):
    """This method is a dummy method that updates all attributes given in ls.

    Parameters
    ----------
    grp : Group
        group to which the attributes belong
    ls : list
        List of string with attributes of group that should be updated in that order
    owner : Frame
        Parent frame object
    args : additional positional arguments
    kwargs : additional keyword arguments

    Notes
    -----
    args and kwargs are only passed to the updater of the Heartbeat, NOT systole or diastole."""
    for val in ls:
        grp.__dict__[val].update(*args, **kwargs)



"""This package contains pre-defined instances of integration schemes that can be used in ``simframe``.
The naming convention is ``<expl/impl>_<order>_<name>(<_additonal>)``.

For example: The fifth-order adaptive Cash-Karp scheme is ``expl_5_cash_karp_adptv``, the 1st-order implicit Euler
scheme using a GMRES solver is ``impl_1_euler_gmres``."""

from simframe.integration.schemes.expl_1_euler import expl_1_euler
from simframe.integration.schemes.expl_2_fehlberg_adptv import expl_2_fehlberg_adptv
from simframe.integration.schemes.expl_2_heun import expl_2_heun
from simframe.integration.schemes.expl_2_heun_euler_adptv import expl_2_heun_euler_adptv
from simframe.integration.schemes.expl_2_midpoint import expl_2_midpoint
from simframe.integration.schemes.expl_2_ralston import expl_2_ralston
from simframe.integration.schemes.expl_3_bogacki_shampine_adptv import expl_3_bogacki_shampine_adptv
from simframe.integration.schemes.expl_3_gottlieb_shu_adptv import expl_3_gottlieb_shu_adptv
from simframe.integration.schemes.expl_3_heun import expl_3_heun
from simframe.integration.schemes.expl_3_kutta import expl_3_kutta
from simframe.integration.schemes.expl_3_ralston import expl_3_ralston
from simframe.integration.schemes.expl_3_ssprk import expl_3_ssprk
from simframe.integration.schemes.expl_4_38rule import expl_4_38rule
from simframe.integration.schemes.expl_4_ralston import expl_4_ralston
from simframe.integration.schemes.expl_4_runge_kutta import expl_4_runge_kutta
from simframe.integration.schemes.expl_5_cash_karp_adptv import expl_5_cash_karp_adptv

from simframe.integration.schemes.impl_1_euler_direct import impl_1_euler_direct
from simframe.integration.schemes.impl_1_euler_gmres import impl_1_euler_gmres
from simframe.integration.schemes.impl_2_midpoint_direct import impl_2_midpoint_direct

from simframe.integration.schemes.update import update

__all__ = ["expl_1_euler",
           "expl_2_fehlberg_adptv",
           "expl_2_heun",
           "expl_2_heun_euler_adptv",
           "expl_2_midpoint",
           "expl_2_ralston",
           "expl_3_bogacki_shampine_adptv",
           "expl_3_gottlieb_shu_adptv",
           "expl_3_heun",
           "expl_3_kutta",
           "expl_3_ralston",
           "expl_3_ssprk",
           "expl_4_38rule",
           "expl_4_ralston",
           "expl_4_runge_kutta",
           "expl_5_cash_karp_adptv",

           "impl_1_euler_direct",
           "impl_1_euler_gmres",
           "impl_2_midpoint_direct",

           "update"
           ]



from simframe import Frame
from simframe import Integrator
from simframe import writers
from simframe.frame import IntVar

import pytest


def test_frame_initialization():
    f = Frame()
    assert isinstance(f, Frame)


def test_frame_writer():
    f = Frame()
    assert f.writer is None
    with pytest.raises(TypeError):
        f.writer = "writer"
    f.writer = writers.hdf5writer
    assert f.writer is writers.hdf5writer
    f = Frame(writer=writers.hdf5writer)
    assert f.writer is writers.hdf5writer


def test_frame_integrator():
    f = Frame()
    assert f.integrator is None
    with pytest.raises(TypeError):
        f.integrator = "test"
    f.integrator = Integrator(IntVar(f, 0.))
    assert isinstance(f.integrator, Integrator)
    f = Frame(integrator=Integrator(IntVar(f, 0.)))
    assert isinstance(f.integrator, Integrator)



"""This package contains infrastructure for solving differential equations within ``simframe``. The ``Integrator`` class
is the basic class that advances the simulation from snapshot to snapshot by executing one integration ``Instruction`` at
a time. Instructions contain a list of integration ``Scheme``. The ``schemes`` package contains pre-defined integration
schemes that are ready to use in ``simframe``."""

from simframe.integration.instruction import Instruction
from simframe.integration.integrator import Integrator
from simframe.integration.scheme import Scheme
import simframe.integration.schemes as schemes

__all__ = ["Instruction",
           "Integrator",
           "Scheme",
           "schemes"]



from simframe.integration.scheme import Scheme

# Butcher coefficients
a10 = 1.0
b0 = 0.5
b1 = 0.5
c1 = 1.0


def _f_expl_2_heun(x0, Y0, dx, *args, dYdx=None, **kwargs):
    """Explicit 2nd-order Heun's method

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    dYdx : Field, optional, default : None
        Current derivative. Will be calculated, if not set.
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated

    Butcher tableau
    ---------------
      0  |  0   0
      1  |  1   0
    -----|---------
         | 1/2 1/2
    """
    k0 = Y0.derivative(x0, Y0) if dYdx is None else dYdx
    k1 = Y0.derivative(x0 + c1*dx, Y0 + a10*k0*dx)

    return dx*(b0*k0 + b1*k1)


expl_2_heun = Scheme(
    _f_expl_2_heun, description="Explicit 2nd-order Heun's method")



from datetime import timedelta
import numpy as np
from time import monotonic

from simframe.frame.group import Group
from simframe.frame.intvar import IntVar

from simframe.integration.integrator import Integrator
from simframe.io.writer import Writer
from simframe.io.progress import Progressbar
from simframe.utils.color import colorize


class Frame(Group):
    """This is the parent object of type ``Group`` that contains all other objects.

    During integration the ``update()`` function of the ``Frame`` object will be called.
    You have to sub-delegete the updates of your other ``Group`` and ``Field`` objects within this function.

    ``Frame`` has additional functionality for writing output files and for integration."""

    __name__ = "Frame"

    _integrator = None
    _verbosity = None
    _progressbar = None
    _writer = None

    def __init__(self, integrator=None, writer=None, updater=None, verbosity=2, progressbar=None, description=""):
        """
        The parent Frame object.

        Parameters
        ----------
        writer : Writer, optional, default : None
            Object of type Writer fir writing output files
        integrator : Integrator, optional, default : None
            Integrator with integration instructions
        updater : Heartbeat, Updater, callable, list or None, optional, default : None
            Updater for updating the frame
        verbosity : int, optional, default : 2
            Level of verbosity
        progressbar : Progresbar or None, optional, default : None
            Progressbar. If None, standard is used
        description : string, optional, default : ""
            Descriptive string of the frame object"""
        super().__init__(self, updater=updater, description=description)
        self.integrator = integrator
        self.progressbar = progressbar
        self.verbosity = verbosity
        self.writer = writer

    @property
    def integrator(self):
        '''``Integrator`` that controls the simulation.'''
        return self._integrator

    @integrator.setter
    def integrator(self, integrator):
        if integrator is not None and type(integrator) is not Integrator:
            raise TypeError("integrator has to be of type Integrator or None.")
        self._integrator = integrator

    @property
    def progressbar(self):
        '''``Progressbar`` for displaying current status.'''
        return self._progressbar

    @progressbar.setter
    def progressbar(self, value):
        if value is None:
            self._progressbar = Progressbar()
        else:
            if not isinstance(value, Progressbar):
                raise TypeError("<progressbar> has to be of type Progressbar.")
            self._progressbar = value

    @property
    def verbosity(self):
        '''Verbosity of the ``Frame`` objects.'''
        return self._verbosity

    @verbosity.setter
    def verbosity(self, value):
        if not isinstance(value, np.int):
            raise TypeError("<verbosity> has to be of type int.")
        self._verbosity = value

    @property
    def writer(self):
        '''``Writer`` object for writing output files.'''
        return self._writer

    @writer.setter
    def writer(self, value):
        if value is not None and not isinstance(value, Writer):
            raise TypeError("writer hat to be of type Writer or None.")
        self._writer = value

    def writeoutput(self, i=0, forceoverwrite=False, filename="", **kwargs):
        """Writes output to file, if ``Writer`` is specified.

        Parameters
        ----------
        i : int, optional, default : 0
            Number of output
        forceoverwrite : boolean, optional, default : False
            If True, this overrules the seetings of the Writer and enforces the file to be overwritten.
        filename : string, optional, default = ""
            if given this will write the output to this file. Otherwise, it uses the standard scheme.
        kwargs : additional keyword arguments
            Additional options that can be passed to the writer"""

        if self.writer is not None:
            self.writer.write(self, i, forceoverwrite, filename, **kwargs)

    def run(self):
        """This method starts the simulation. An ``Integrator`` has to be set beforehand."""

        if not isinstance(self.integrator, Integrator):
            raise RuntimeError("No integrator set.")

        # Check if integration variable is set
        if not isinstance(self.integrator.var, IntVar):
            raise RuntimeError(
                "No integration variable assigned to integrator.")

        # If there are no snapshots set
        if not len(self.integrator.var.snapshots):
            raise RuntimeError(
                "No snapshots set. At least one snapshot has to be given.")

        # If integration variable passed maximum value of snapshots
        if self.integrator.var >= self.integrator.var.snapshots[-1]:
            raise RuntimeError(
                "Integration variable already passed the largest snapshot.")

        # Timekeeping
        tini = monotonic()

        # Write initial conditions
        if self.integrator.var < self.integrator.var.snapshots[0]:
            self.writeoutput(0)

        # Staring index of snapshots
        starting_index = np.argmin(
            self.integrator.var >= self.integrator.var.snapshots)
        # Starting value of integration variable
        startingvalue = self.integrator.var.copy()
        for i in range(starting_index, len(self.integrator.var.snapshots)):

            # Nextsnapshot cannot be referenced directly, because it dynamically changes.
            nextsnapshot = self.integrator.var.nextsnapshot
            prevsnapshot = self.integrator.var.prevsnapshot if self.integrator.var.prevsnapshot is not None else startingvalue

            while self.integrator.var < nextsnapshot:

                if self.verbosity > 1:
                    self.progressbar(self.integrator.var,
                                     prevsnapshot,
                                     nextsnapshot,
                                     startingvalue,
                                     self.integrator.var.snapshots[-1])

                self.integrator.integrate()
                self.integrator.var += self.integrator.var.stepsize
                self.update()

            if self.verbosity > 1:
                self.progressbar._reset()

            self.writeoutput(i + 1)

        # Timekeeping
        tfin = monotonic()
        t_exec = timedelta(seconds=np.int(tfin-tini))
        if self.verbosity > 0:
            msg = "Execution time: {}".format(colorize(t_exec, color="blue"))
            print(msg)



"""This package is for input/output operations. It contains template ``Writer`` and ``Reader`` classes that can be used
to create customized writing and reading methods. The package ``writers`` contains pre-defined ``Writer`` instances for
writing and reading ``simframe`` data. The package furthermore contains a method for reading dump files and for printing
a progress bar in an interactive shell."""

from simframe.io.reader import Reader
from simframe.io.writer import Writer
from simframe.io import writers
from simframe.io.dump import readdump
from simframe.io.progress import Progressbar

__all__ = ["Reader",
           "Writer",
           "writers",
           "readdump",
           "Progressbar"]



from simframe.frame.abstractgroup import AbstractGroup


class Scheme:
    """Class for an integration ``Scheme`` that can be used as template for creating custom schemes.

    Notes
    -----
    The integration ``Scheme`` needs to return ``False`` if the integration failed. The ``Integrator`` will then
    perform a fail operation and will try it again. This can be used to implement schemes with adaptive
    step sizes. If the step size was not small enough the fail operation can reduce it further."""

    __name__ = "Scheme"

    _description = ""
    _controller = {}
    _scheme = None

    def __init__(self, scheme, controller={}, description=""):
        """Abstract integration scheme.
        The integration scheme itself is callable.

        Parameters
        ----------
        scheme : callable
            Function that returns the delta of the variable to be integrated
        controller : dict, optional, default : {}
            Additional keyword arguments passed to integration scheme
        description : string, optional, default : ""
            Descriptive string of the integration scheme"""
        self.scheme = scheme
        self.controller = controller
        self.description = description

    @property
    def scheme(self):
        '''Integration ``Scheme``.'''
        return self._scheme

    @scheme.setter
    def scheme(self, value):
        if not hasattr(value, "__call__"):
            raise TypeError("Scheme function needs to be callable.")
        self._scheme = value

    @property
    def controller(self):
        '''Dictionary with keyword arguments that is passed to the integration ``Scheme``.'''
        return self._controller

    @controller.setter
    def controller(self, value):
        if not isinstance(value, dict):
            raise TypeError("controller has to be of type dict.")
        self._controller = value

    @property
    def description(self):
        '''Description of the ``Scheme``.'''
        return self._description

    @description.setter
    def description(self, value):
        if not isinstance(value, str):
            raise ValueError("<value> has to be of type str.")
        self._description = value

    def __call__(self, x0, Y0, dx, *args, **kwargs):
        """Method for returning the new value of the variable to be integrated.

        Parameters
        ----------
        x0 : Intvar
            Integration variable at beginning of scheme
        Y0 : Field
            Variable to be integrated at the beginning of scheme
        dx : IntVar
            Stepsize of integration variable
        controller : dict, optional, default : {}
            Additional keyword arguments passed to integration scheme
        args : additional positional arguments
        kwargs : additional kewyowrd arguments


        Returns
        -------
        Y1 : Field or False
            New value of the variable to be integrated.
            Functions needs to return ``False`` if integration failed."""
        return self.scheme(x0, Y0, dx, *args, **kwargs)

    def __str__(self):
        return AbstractGroup.__str__(self)

    def __repr__(self):
        return self.__str__()



import glob
import h5py
import numbers
import numpy as np
from types import SimpleNamespace
import os

from simframe.io.reader import Reader
from simframe.io.writer import Writer
from simframe.frame.field import Field


def _hdf5wrapper(obj, filename, com="lzf", comopts=None):
    """Wrapper to write object to HDF5 file.

    This function recursively calls a another functions thats goes through the object tree.

    Parameters
    ----------
    obj : object
        the object to be stored in a file
    filename : string
        path to file

    Keywords
    --------
    com : string
        compression method to be used by `h5py`
    comopt : compression_opts
        compression options, see `h5py.File`'s `create_dataset` for details
    """

    with h5py.File(filename, "w") as hdf5file:
        _writehdf5(obj, hdf5file, com=com, comopts=comopts)


def _writehdf5(obj, file, com="lzf", comopts=None, prefix=""):
    """Writes a given object to a h5py file.

    By default all attributes of the object are written out, excluding the ones that start with an underscore.
    Fields with attribute Field.save == False will be skipped.

    Parameters:
    ----------
    obj : object
        the object to be stored in a file
    file : hdf5 file
        open hdf5 file object

    Keywords
    --------
    com : string
        compression method to be used by `h5py`
    comopt : compression_opts
        compression options, see `h5py.File`'s `create_dataset` for details
    prefix : str
        a prefix prepended to the name of each attribute when storing with h5py
    """

    if hasattr(obj, "_description") and obj._description is not None and prefix == "":
        file.create_dataset(
            "description",
            data=obj._description
        )

    for key, val in obj.__dict__.items():

        # Ignore hidden variables
        if key.startswith('_'):
            continue
        # Skip fields that should not be stored
        if isinstance(val, Field) and val.save == False:
            continue

        name = prefix + key

        # Check if numpy.ndarray of strings and convert to list
        if isinstance(val, np.ndarray) and val.dtype.type is np.str_:
            val = val.tolist()

        # Check for number
        if isinstance(val, (numbers.Number, np.number)):
            file.create_dataset(
                name,
                data=val
            )
        # Check for tuple/list/dict
        elif type(val) in [tuple, list, dict]:
            # special case for list of strings
            if all([type(_v) == str for _v in val]):
                file.create_dataset(
                    name,
                    data=np.array(val, dtype=object),
                    dtype=h5py.special_dtype(vlen=str),
                    compression=com,
                    compression_opts=comopts)
            else:
                file.create_dataset(
                    name,
                    data=val,
                    compression=com,
                    compression_opts=comopts
                )
        # Check for string
        elif type(val) is str:
            file.create_dataset(
                name,
                data=val
            )
        # Check for Numpy array
        elif isinstance(val, np.ndarray):
            if val.shape == ():
                file.create_dataset(
                    name,
                    data=val,
                )
            else:
                file.create_dataset(
                    name,
                    data=val,
                    compression=com,
                    compression_opts=comopts
                )
        # Check for None
        elif val is None:
            file.create_dataset(
                name,
                data=0
            )
        # Other objects
        else:
            _writehdf5(val, file, com=com,
                       comopts=comopts, prefix=name + "/")


class hdf5reader(Reader):
    """Reader class for the HDF5 writer."""

    def __init__(self, writer):
        """HDF5 reader

        Parameters
        ----------
        writer : Writer
            Writer object to which the reaer belongs."""
        super().__init__(writer)

    def output(self, filename):
        """Reads a single output file.

        Parameters
        ----------
        filename : str
            Path to filename to be read

        Returns
        -------
        data : SimpleNamespace
            Namespace of data in file."""

        if not isinstance(filename, str):
            filename = self._writer._getfilename(filename)

        if not os.path.isfile(filename):
            raise RuntimeError("File does not exist.")

        with h5py.File(filename, "r") as hdf5file:
            return self._readgroup(hdf5file)

    def sequence(self, field):
        """Reading the entire sequence of a specific field.

        Parameters
        ----------
        field : string
            String with location of requested field

        Returns
        -------
        seq : array
            Array with requested values

        Notes
        -----
        ``field`` is addressing the values just as in the parent frame object.
        E.g. ``"groupA.groupB.fieldC"`` is addressing ``Frame.groupA.groupB.fieldC``."""
        files = self.listfiles()
        if files == []:
            raise RuntimeError("<datadir> does not exist or is empty.")
        if not isinstance(field, str):
            raise TypeError("<field> has to be of type string.")
        loc = field.replace(".", "/")
        ret = []
        for f in files:
            with h5py.File(f, "r") as hdf5file:
                A = np.array(hdf5file[loc][()])
                if A.shape == (1,):
                    ret.append(A[0])
                else:
                    ret.append(A)
        return np.array(ret)

    def _readgroup(self, gr):
        """Helper function that is iteratively called to get the depth of the data set.

        Parameters
        ----------
        gr : Group of type h5py._hl.group.Group
            The h5py data set to be read

        Returns
        -------
        data : SimpleNamespace
            Namespace of data"""
        ret = {}
        for ds in gr.keys():
            if isinstance(gr[ds], h5py._hl.group.Group):
                ret[ds] = self._readgroup(gr[ds])
            else:
                ret[ds] = gr[ds][()]
        return SimpleNamespace(**ret)


hdf5writer = Writer(_hdf5wrapper, extension="hdf5", description="HDF5 file format using h5py", options={
                    "com": "lzf", "comopts": None}, reader=hdf5reader)



"""This package contains pre-defined ``Writer`` instances that can be used for writing and reading ``Frame`` objects.
The ``hdf5writer`` writes data files in the HDF5 file format. The ``namespacewriter`` does not write output files
(except for dump files if required). The data is stored locally in the ``Writer`` object itself."""

from simframe.io.writers.hdf5writer import hdf5writer
from simframe.io.writers.namespacewriter import namespacewriter

__all__ = ["hdf5writer",
           "namespacewriter"]



from simframe.integration.scheme import Scheme

import numpy as np

# Butcher coefficients
a10 = 1/2
a21 = 3/4
a30, a31, a32 = 2/9, 1/3, 4/9
b0, b1, b2 = 2/9, 1/3, 4/9
bs0, bs1, bs2, bs3 = 7/24, 1/4, 1/3, 1/8
e0, e1, e2, e3 = b0-bs0, b1-bs1, b2-bs2, -bs3
c1, c2 = 1/2, 3/4


def _f_expl_3_bogacki_shampine_adptv(x0, Y0, dx, *args, dYdx=None, econ=0.005832, eps=0.1, pgrow=-1/3, pshrink=-0.5, safety=0.9, **kwargs):
    """Explicit adaptive 3rd-order Bogacki-Shampine method

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    dYdx : Field, optional, default : None
        Current derivative. Will be calculated, if not set.
    econ : float, optional, default : 0.005832
        Error controll parameter for setting stepsize
    eps : float, optional, default : 0.9
        Desired maximum relative error
    prgrow : float, optional, default : -1/3
        Power for increasing step size
    pshrink : float, optional, default : -1/2
        Power for decreasing stepsize
    safety : float, optional, default : 0.9
        Safety factor when changing step size
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated
        False if step size too large

    Butcher tableau
    ---------------
      0  |  0    0   0   0
     1/2 | 1/2   0   0   0
     3/4 |  0   3/4  0   0
      1  | 2/9  1/3 4/9  0
    -----|------------------
         | 2/9  1/3 4/9  0
         | 7/24 1/4 1/3 1/8
    """
    k0 = Y0.derivative(x0, Y0) if dYdx is None else dYdx
    k1 = Y0.derivative(x0 + c1*dx, Y0 + a10*k0 * dx)
    k2 = Y0.derivative(x0 + c2*dx, Y0 + a21*k1 * dx)
    k3 = Y0.derivative(x0 + dx, Y0 + (a30*k0 + a31*k1 + a32*k2)*dx)

    Yscale = np.abs(Y0) + np.abs(dx*k0)
    Yscale[Yscale == 0.] = 1.e100       # Deactivate for zero crossings

    e = dx*(e0*k0 + e1*k1 + e2*k2 + e3*k3)
    emax = np.max(np.abs(e/Yscale)) / eps

    # Integration successful
    if emax <= 1.:
        # Suggest new stepsize
        dxnew = safety*dx*emax**pgrow if econ < emax else 5.*dx
        x0.suggest(dxnew)
        return dx*(b0*k0 + b1*k1 + b2*k2)
    else:
        # Suggest new stepsize
        dxnew = np.maximum(safety*dx*emax**pshrink, 0.1*dx)
        x0.suggest(dxnew)
        return False


expl_3_bogacki_shampine_adptv = Scheme(
    _f_expl_3_bogacki_shampine_adptv, description="Explicit adaptive 3rd-order Bogacki-Shampine method")



from simframe.integration.scheme import Scheme

import numpy as np

# Butcher coefficients
b0, b1 = 1/2, 1/2
e0, e1 = -1/2, 1/2


def _f_expl_2_heun_euler_adptv(x0, Y0, dx, *args, dYdx=None, econ=0.0324, eps=0.1, pgrow=-0.5, pshrink=-1., safety=0.9, **kwargs):
    """Explicit adaptive 2nd-order Heun-Euler method

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    dYdx : Field, optional, default : None
        Current derivative. Will be calculated, if not set.
    econ : float, optional, default : 0.0324
        Error controll parameter for setting stepsize
    eps : float, optional, default : 0.9
        Desired maximum relative error
    prgrow : float, optional, default : -0.5
        Power for increasing step size
    pshrink : float, optional, default : -1.
        Power for decreasing stepsize
    safety : float, optional, default : 0.9
        Safety factor when changing step size
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated
        False if step size too large

    Butcher tableau
    ---------------
      0  |  0   0
      1  |  1   0
    -----|---------
         | 1/2 1/2
         |  1   0
    """
    k0 = Y0.derivative(x0, Y0) if dYdx is None else dYdx
    k1 = Y0.derivative(x0 + dx, Y0 + k0*dx)

    Yscale = np.abs(Y0) + np.abs(dx*k0)
    Yscale[Yscale == 0.] = 1.e100       # Deactivate for zero crossings

    e = dx*(e0*k0 + e1*k1)
    emax = np.max(np.abs(e/Yscale)) / eps

    # Integration successful
    if emax <= 1.:
        # Suggest new stepsize
        dxnew = safety*dx*emax**pgrow if econ < emax else 5.*dx
        x0.suggest(dxnew)
        return dx*(b0*k0 + b1*k1)
    else:
        # Suggest new stepsize
        dxnew = np.maximum(safety*dx*emax**pshrink, 0.1*dx)
        x0.suggest(dxnew)
        return False


expl_2_heun_euler_adptv = Scheme(
    _f_expl_2_heun_euler_adptv, description="Explicit adaptive 2nd-order Heun-Euler method")



from simframe.integration.scheme import Scheme

# Butcher coefficients
a10 = 1/2
a21 = 1/2
b0, b1, b2, b3 = 1/6, 1/3, 1/3, 1/6
c1, c2 = 1/2, 1/2


def _f_expl_4_runge_kutta(x0, Y0, dx, *args, dYdx=None, **kwargs):
    """Explicit 4th-order classical Runge-Kutta method

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    dYdx : Field, optional, default : None
        Current derivative. Will be calculated, if not set.
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated

    Butcher tableau
    ---------------
      0  |  0   0   0   0
     1/2 | 1/2  0   0   0
     1/2 |  0  1/2  0   0
      1  |  0   0   1   0
    -----|-----------------
         | 1/6 1/3 1/3 1/6
    """

    k0 = Y0.derivative(x0, Y0) if dYdx is None else dYdx
    k1 = Y0.derivative(x0 + c1*dx, Y0 + a10*k0*dx)
    k2 = Y0.derivative(x0 + c2*dx, Y0 + a21*k1*dx)
    k3 = Y0.derivative(x0 + dx, Y0 + k2*dx)

    return dx*(b0*k0 + b1*k1 + b2*k2 + b3*k3)


expl_4_runge_kutta = Scheme(
    _f_expl_4_runge_kutta, description="Explicit 4th-order classical Runge-Kutta method")



"""Package contains utility classes that facilitate the use of ``simframe``

``Color`` is a generic class that can be used to colorize text. ``colorize`` is an instance of ``Color``, that can
be called to add decorators to a string for colored output."""

from simframe.utils.color import Color
from simframe.utils.color import colorize

__all__ = ["Color",
           "colorize"]



from simframe.integration.scheme import Scheme

# Butcher coefficients
a10 = 0.5
c1 = 0.5


def _f_expl_2_midpoint(x0, Y0, dx, *args, dYdx=None, **kwargs):
    """Explicit 2nd-order midpoint method

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    dYdx : Field, optional, default : None
        Current derivative. Will be calculated, if not set.
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated

    Butcher tableau
    ---------------
      0  |  0   0
     1/2 | 1/2  0
    -----|---------
         |  0   1
    """
    k0 = Y0.derivative(x0, Y0) if dYdx is None else dYdx
    k1 = Y0.derivative(x0 + c1*dx, Y0 + a10*k0*dx)

    return dx*k1


expl_2_midpoint = Scheme(
    _f_expl_2_midpoint, description="Explicit 2nd-order midpoint method")



from simframe.integration.scheme import Scheme


def _update(x0, Y0, dx, *args, **kwargs):
    """Instruction for updating Y

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    args : additional positional arguments
    kwargs : additional keyword arguments

    Returns
    -------
    dY : True"""
    if Y0.buffer is not None:
        Y0._setvalue(Y0 + Y0._buffer)
        Y0._buffer = None
    return True


update = Scheme(_update, description="Instruction to update Y")



from simframe.integration.scheme import Scheme

# Butcher coefficients
a10 = 1/3
a20 = -1/3
b0, b1, b2, b3 = 1/8, 3/8, 3/8, 1/8
c1, c2 = 1/3, 2/3


def _f_expl_4_38rule(x0, Y0, dx, *args, dYdx=None, **kwargs):
    """Explicit 4th-order 3/8 rule method

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    dYdx : Field, optional, default : None
        Current derivative. Will be calculated, if not set.
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated

    Butcher tableau
    ---------------
      0  |  0   0   0   0
     1/3 | 1/3  0   0   0
     2/3 |-1/3  1   0   0
      1  |  1  -1   1   0
    -----|-----------------
         | 1/8 3/8 3/8 1/8
    """

    k0 = Y0.derivative(x0, Y0) if dYdx is None else dYdx
    k1 = Y0.derivative(x0 + c1*dx, Y0 + a10*k0 * dx)
    k2 = Y0.derivative(x0 + c2*dx, Y0 + (a20*k0 + k1)*dx)
    k3 = Y0.derivative(x0 + dx, Y0 + (k0 - k1 + k2)*dx)

    return dx*(b0*k0 + b1*k1 + b2*k2 + b3*k3)


expl_4_38rule = Scheme(
    _f_expl_4_38rule, description="Explicit 4th-order 3/8 rule method")



from simframe.integration.scheme import Scheme

# Butcher coefficients
a20, a21 = 1/4, 1/4
b0, b1, b2 = 1/6, 1/6, 2/3
c2 = 1/2


def _f_expl_3_ssprk(x0, Y0, dx, *args, dYdx=None, **kwargs):
    """Explicit 3rd-order Strong Stability Preserving Runge-Kutta method

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    dYdx : Field, optional, default : None
        Current derivative. Will be calculated, if not set.
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated

    Butcher tableau
    ---------------
      0  |  0   0   0
      1  |  1   0   0
     1/2 | 1/4 1/4  0
    -----|-------------
         | 1/6 1/6 2/3
    """
    k0 = Y0.derivative(x0, Y0) if dYdx is None else dYdx
    k1 = Y0.derivative(x0 + dx, Y0 + k0 * dx)
    k2 = Y0.derivative(x0 + c2*dx, Y0 + (a20*k0 + a21*k1)*dx)

    return dx*(b0*k0 + b1*k1 + b2*k2)


expl_3_ssprk = Scheme(
    _f_expl_3_ssprk, description="Explicit 3rd-order Strong Stability Preserving Runge-Kutta method")



from simframe.frame.abstractgroup import AbstractGroup
from simframe.frame.field import Field
from simframe.frame.group import Group
from simframe.io.reader import Reader
from simframe.io.writer import Writer
from simframe.utils.color import colorize

from collections import deque
import copy
import numbers
import numpy as np
from types import SimpleNamespace


class _namespacewriter(Writer):
    """Modified class to write ``Frame`` object to namespace"""

    _buffer = deque([])

    def __init__(self, func, datadir="data", zfill=0, dumping=True, reader=None, description="Temporary namespace writer"):
        """
        Parameters
        ----------
        func : callable
            Function that writes Frame object.
        datadir : str, optional, default : "data"
            Path to data directory. Only used for dump files
        zfill : int, optional default : 4
            leading zeros of output numbers
        dumping : boolean, optional, default : True
            if True dump files are written
        reader, Reader, optional, default : None
            Reader to read the outputs
        description : str, optional, default : "Namespace writer"
        """
        super().__init__(func, datadir=datadir, filename="_", zfill=zfill, extension="",
                         overwrite=False, dumping=dumping, reader=reader, description=description)

    def __repr__(self):
        ret = self.__str__()+"\n"
        ret += "-" * (len(ret)-1) + "\n"
        ret += "    Data directory : {}\n".format(self.datadir)
        ret += "    Dumping        : {}\n".format(
            colorize(self.dumping, "yellow") if not self.dumping else self.dumping)
        ret += "    Verbosity      : {}".format(self.verbosity)
        return ret

    def _getfilename(self):
        """Filenames are not required for this class."""
        pass

    def write(self, owner, i=0, forceoverwrite=False, filename=""):
        """Writes output to namespace

        Parameters
        ----------
        owner : Frame
            Parent frame object
        i : int
            Not used in this class
        forceoverwrite : boolean
            Not used in this class
        filename : string
            Not used in this class"""
        self._buffer.append(self._func(owner))
        if self.verbosity > 0:
            num = str(i).zfill(self._zfill)
            msg = "Saving frame {}".format(num)
            print(msg)
        if self.dumping:
            self.writedump(owner)

    def reset(self):
        """This resets the namespace.

        Notes
        -----
        WARNING: This cannot be undone."""
        self._buffer.clear()


class namespacereader(Reader):
    """Class to read namespace outputs"""

    def __init__(self, writer):
        super().__init__(writer)

    def all(self):
        """Functions that reads all output files and combines them into a single ``SimpleNamespace``.

        Returns
        -------
        dataset : SimpleNamespace
            Namespace of data set.

        Notes
        -----
        This function is reading one output files to get the structure of the data and
        calls ``read.sequence()`` for every field in the data structure."""
        if self._writer._buffer == deque([]):
            raise RuntimeError("Writer buffer is empty.")
        # Read first file to get structure
        data0 = self._writer._buffer[0]
        return self._expand(data0)

    def output(self, i):
        """Reading a single output

        Parameters
        ----------
        i : int
            Number of output to be read

        Returns
        -------
        n : SimpleNamespace
            Namespace of desired output"""
        if i >= len(self._writer._buffer):
            raise RuntimeError("Output {} does not exist.".format(i))

        return self._writer._buffer[i]

    def sequence(self, field):
        """Reading the entire sequence of a specific field.

        Parameters
        ----------
        field : string
            String with location of requested field

        Returns
        -------
        seq : array
            Array with requested values

        Notes
        -----
        ``field`` is addressing the values just as in the parent frame object.
        E.g. ``"groupA.groupB.fieldC"`` is addressing ``Frame.groupA.groupB.fieldC``."""
        if self._writer._buffer == deque([]):
            raise RuntimeError("Writer buffer is empty.")
        if not isinstance(field, str):
            raise TypeError("<field> has to be string.")
        loc = field.split(".")
        N = len(self._writer._buffer)
        ret = []
        for i in range(N):
            A = np.array(_getvaluefrombuffer(self._writer._buffer[i], loc))
            if A.shape == (1,):
                ret.append(A[0])
            else:
                ret.append(A)
        return np.array(ret)


def _getvaluefrombuffer(buf, loc):
    """Returns a requested value from buffer.
    Function is called recursively.

    Parameters
    ----------
    buf : dict
        Buffer object
    loc : list
        List of strings with the requested location within buf

    Returns
    -------
    ret : object
        Reqested value in buf at position loc"""
    if len(loc) > 1:
        return _getvaluefrombuffer(buf.__dict__[loc[0]], loc[1:])
    if not hasattr(buf, loc[0]):
        raise KeyError("Requested <field> does not exist.")
    return buf.__dict__[loc[0]]


def _converttonamespace(o):
    """Converts an object into a namespace

    Parameters
    ----------
    o : object
        object

    Returns
    -------
    ns : SimpleNamespace
        Nested namespace with the data in Frame object

    Notes
    -----
    Attributes beginning with underscore _ are being ignored.
    So are fields with Field.save == False."""
    ret = {}

    # These things are written directy into the dictionary.
    direct = (numbers.Number, np.number, tuple,
              list, np.ndarray, str)

    for key, val in o.__dict__.items():

        # Ignore hidden variables
        if key.startswith("_"):
            continue
        # Skip fields that should not be stored
        if isinstance(val, Field) and val.save == False:
            continue

        if val is not None and isinstance(val, direct):
            ret[key] = copy.copy(val)
        else:
            ret[key] = _converttonamespace(val)

    return SimpleNamespace(**ret)


def _writeframetonamespace(frame):
    """Takes a list of dicts and a frame, stitches them together and returns namespace.

    Paramters
    ---------
    frame : Frame
        Frame object to add

    Returns
    -------
    n : SimpleNamespace
        Namespace with data"""
    d = _converttonamespace(frame)
    return d


namespacewriter = _namespacewriter(
    _writeframetonamespace, reader=namespacereader, description="Namespace writer")



from collections import deque
import glob
import numpy as np
from types import SimpleNamespace
import os

from simframe.frame.abstractgroup import AbstractGroup


class Reader(object):
    """General class for reading output files.
    Every ``Writer`` class should also provide a reader for its data files.

    Custom ``Reader`` must provide a method ``Reader.output()`` that reads a single output file
    and returns the data set as type ``SimpleNamespace``.

    The general ``Reader`` class provides a function that stitches together all ``SimpleNamespaces``
    the ``Reader.output()`` method provides into a single ``SimpleNamespace`` by adding another dimension
    along the integration variable ``IntVar``."""

    __name__ = "Reader"

    _description = ""
    _writer = None

    def __init__(self, writer, description=""):
        """General ``Reader`` class

        Parameters
        ----------
        writer : Writer
            The writer object to which the reader belongs
        description : str, optional, default = ""
            Descriptive string of reader."""
        self.description = description
        self._writer = writer

    @property
    def description(self):
        '''Description of the ``Reader``.'''
        return self._description

    @description.setter
    def description(self, value):
        if not isinstance(value, (str, type(None))):
            raise ValueError("<value> has to be of type str.")
        self._description = value

    def __str__(self):
        return AbstractGroup.__str__(self)

    def __repr__(self):
        return self.__str__()

    def output(self, file):
        """Function that returns the data of a single output file.

        Parameters
        ----------
        file : str
            Path to file that should be read.

        Returns
        -------
        data : SimpleNamespace
            Data set of a single output file."""
        raise NotImplementedError("<read.output> is not implemented.")

    def sequence(self, field):
        """Function that returns the entire sequence of a specific field.

        Parameters
        ----------
        field : str
            String with location of requested field

        Returns
        -------
        seq : array
            Array with requested values

        Notes
        -----
        ``field`` is addressing the values just as in the parent frame object.
        E.g. ``"groupA.groupB.fieldC"`` is addressing ``Frame.groupA.groupB.fieldC``."""
        raise NotImplementedError("<read.sequence> is not implemented.")

    def listfiles(self):
        """Method to list all data files in a directory

        Returns
        -------
        files : list
            List of strings of all found data files sorted alphanumerically.

        Notes
        -----
        Function only searches for files that match the pattern specified by the ``Writer``'s
        ``filename`` and ``extension`` attributes."""
        datadir = self._writer.datadir
        ext = self._writer.extension if self._writer.extension != "" else "." + \
            self._writer.extension
        wildcard = os.path.join(datadir, self._writer.filename + "*" + ext)
        files = glob.glob(wildcard)
        files = sorted(files, key=str.casefold)
        return files

    def all(self):
        """Functions that reads all output files and combines them into a single ``SimpleNamespace``.

        Returns
        -------
        dataset : SimpleNamespace
            Namespace of data set.

        Notes
        -----
        This function is reading one output files to get the structure of the data and
        calls ``read.sequence()`` for every field in the data structure."""
        files = self.listfiles()
        if files == []:
            raise RuntimeError("Data directory does not exist or is empty.")
        # Read first file to get structure
        data0 = self.output(files[0])
        # print(data0.__dict__.keys())
        return self._expand(data0)

    def _expand(self, ns, prefix=""):
        """This is a function that get recursively called to fill a data structure with sequences.

        Parameters
        ----------
        ns : SimpleNamespace
            Name space to read sequences from
        prefix : string, optional default: ""
            prefix of data to get into depth of the structure"""
        ret = {}
        for key, val in ns.__dict__.items():
            new_prefix = ".".join(filter(None, [prefix, key]))
            if isinstance(val, SimpleNamespace):
                ret[key] = self._expand(val, prefix=new_prefix)
            else:
                ret[key] = self.sequence(new_prefix)
        return SimpleNamespace(**ret)



from simframe.integration.scheme import Scheme

import numpy as np

# Butcher coefficients
a10 = 1/2
a20, a21 = 3/40, 9/40
a30, a31, a32 = 3/10, -9/10, 6/5
a40, a41, a42, a43 = -11/54, 5/2, -70/27, 35/27
a50, a51, a52, a53, a54 = 1631/55296, 175 / \
    512, 575/13824, 44275/110592, 253/4096
b0, b2, b3, b5 = 37/378, 250/621, 125/594, 512/1771
bs0, bs2, bs3, bs4, bs5 = 2825/27648, 18575/48384, 13525/55296, 277/14336, 1/4
e0, e2, e3, e4, e5 = b0-bs0, b2-bs2, b3-bs3, -bs4, b5-bs5
c1, c2, c3, c5 = 1/5, 3/10, 3/10, 7/8


def _f_expl_5_cash_karp_adptv(x0, Y0, dx, *args, dYdx=None, econ=0.0001889568, eps=0.1, pgrow=-0.2, pshrink=-0.25, safety=0.9, **kwargs):
    """Explicit adaptive 5th-order Cash-Karp method

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    dYdx : Field, optional, default : None
        Current derivative. Will be calculated, if not set.
    econ : float, optional, default : 0.005832
        Error controll parameter for setting stepsize
    eps : float, optional, default : 0.9
        Desired maximum relative error
    prgrow : float, optional, default : -1/3
        Power for increasing step size
    pshrink : float, optional, default : -1/2
        Power for decreasing stepsize
    safety : float, optional, default : 0.9
        Safety factor when changing step size
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated
        False if step size too large

    Butcher tableau
    ---------------
      0   |     0         0         0           0          0         0
     1/5  |    1/5        0         0           0          0         0
     3/10 |    3/40      9/40       0           0          0         0
     3/5  |    3/10     -9/10      6/5          0          0         0
      1   |  -11/54      5/2     -70/27       35/27        0         0
     7/8  | 1631/55296 175/512   575/13824 44275/110592 253/4096     0
    ------|----------------------------------------------------------------
          |   37/378      0      250/621     125/594       0      512/1771
          | 2825/27648    0    18575/48384 13525/55296  277/14336   1/4
    """
    k0 = Y0.derivative(x0, Y0) if dYdx is None else dYdx
    k1 = Y0.derivative(x0 + c1*dx, Y0 + a10*k0 * dx)
    k2 = Y0.derivative(x0 + c2*dx, Y0 + (a20*k0 + a21*k1)*dx)
    k3 = Y0.derivative(x0 + c3*dx, Y0 + (a30*k0 + a31*k1 + a32*k2)*dx)
    k4 = Y0.derivative(x0 + dx, Y0 + (a40*k0 + a41*k1 + a42*k2 + a43*k3)*dx)
    k5 = Y0.derivative(x0 + c5*dx, Y0 + (a50*k0 + a51 *
                                         k1 + a52*k2 + a53*k3 + a54*k4)*dx)

    Yscale = np.abs(Y0) + np.abs(dx*k0)
    Yscale[Yscale == 0.] = 1.e100       # Deactivate for zero crossings

    e = dx*(e0*k0 + e2*k2 + e3*k3 + e4*k4 + e5*k5)
    emax = np.max(np.abs(e/Yscale)) / eps

    # Integration successful
    if emax <= 1.:
        # Suggest new stepsize
        dxnew = safety*dx*emax**pgrow if econ < emax else 5.*dx
        x0.suggest(dxnew)
        return dx*(b0*k0 + b2*k2 + b3*k3 + b5*k5)
    else:
        # Suggest new stepsize
        dxnew = np.maximum(safety*dx*emax**pshrink, 0.1*dx)
        x0.suggest(dxnew)
        return False


expl_5_cash_karp_adptv = Scheme(
    _f_expl_5_cash_karp_adptv, description="Explicit adaptive 5th-order Cash-Karp method")



from simframe.frame.heartbeat import Heartbeat


class AbstractGroup(object):
    """This is an abstract class that serves as template for other classes.

    ``AbstractGroup`` has a descriptive string, an owner and an updater.
    The owner is the parent ``Frame`` object and is hidden. The updater is of type ``Heartbeat``.

    ``AbstractGroup`` has an ``update`` method that is calling ``systole``, ``updater``, and ``diastole``
    of the ``Heartbeat`` object, which manages the update of ``AbstractGroup``.

    ``AbstractGroup`` should not be instanciated directly."""

    __name__ = ""

    _description = ""
    _owner = None
    _updater = Heartbeat(None)

    @property
    def description(self):
        '''Description of the instance.'''
        return self._description

    @description.setter
    def description(self, value):
        if not isinstance(value, str):
            raise ValueError("<value> has to be of type str.")
        self._description = value

    @property
    def updater(self):
        '''``Heatbeat`` object with instructions for updating the instance.'''
        return self._updater

    @updater.setter
    def updater(self, value):
        if isinstance(value, Heartbeat):
            self._updater = value
        else:
            self._updater = Heartbeat(value)

    def __str__(self):
        ret = "{}".format(str(self.__name__))
        description = " (" + self.description + \
            ")" if self.description != "" else self.description
        ret = "{}{}".format(self.__name__, description)
        return ret

    def __repr__(self):
        return self.__str__()

    def update(self, *args, **kwargs):
        """Function to update the object.
        Ths functions calls the heartbeat instance of the object.

        Parameters
        ----------
        args : additional positional arguments
        kwargs : additional keyword arguments

        Notes
        -----
        Positional arguments and keyword arguments are only passed to the ``updater``,
        NOT to ``systole`` and ``diastole``."""
        self.updater.beat(self._owner, *args, **kwargs)



import numpy as np

from simframe.frame.field import Field
from simframe.frame.heartbeat import Heartbeat
from simframe.utils.color import colorize


class IntVar(Field):
    """Cclass for integration variables that behaves as ``Field`` but has additional functionality with respect to
    stepsize management for integration.

    Notes
    -----
    The ``updater`` for integration variables is calculating the stepsize. The function associated to the
    ``updater`` needs the parent ``Frame`` object as first positional argument and needs to return the
    desired stepsize.

    ``IntVar.update()`` does not update the integration variable. Try not to update the integration variable by hand.
    Let the ``Integrator`` do it for you."""

    _snapshots = []
    _suggested = None
    _prevstepsize = 0.

    def __new__(cls, owner, value=0, snapshots=[], updater=None, description="", save=True, copy=False):
        """Parameters
        ----------
        owner : Frame
            Parent frame object to which the field belongs
        value : number, optional, default : 0
            Initial value of the field
        snapshots : list, array, optional, default : []
            List or array of snapshots at which an output file should be written.
            Has to be monotonously increasing.
        updater : Heartbeat, Updater, callable, optional, default : None
            Updater for calculating stepsize
        description : string, optional, default : ""
            Descriptive string for the field
        save : boolean, optional, default : True
            If False the integration variable is not written into output files
        copy : boolean, optional, default : True
            If True <value> will be copied, not referenced"""
        obj = super().__new__(cls, owner, value, updater=updater,
                              description=description, constant=False, save=save, copy=copy)
        obj.snapshots = snapshots
        obj._prevstepsize = 0.
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        super().__array_finalize__(obj)
        self.snapshots = getattr(obj, "snapshots", [])
        self._prevstepsize = 0.

    def __str__(self):
        ret = "{}".format(str(self.__name__))
        ret = super().__str__()
        ret += ", {}".format(colorize("Integration variable", "purple"))
        return ret

    def update(self):
        """Not used for ``IntVar``."""
        msg = "{}: {}".format(colorize("Warning", "yellow"),
                              "Do not update the integration variable by hand.")
        print(msg)

    def _update(self, u, *args, upd=False, **kwargs):
        msg = "{}: {}".format(colorize("Warning", "yellow"),
                              "Do not update the integration variable by hand.")
        print(msg)

    def suggest(self, value, reset=False):
        """Suggest a step size

        For adaptive integration schemes, this function can be used to suggest a step size for the next
        integration step. If many vaiables are integrated this safes the smallest suggested step size
        in a temporary buffer accessible via ``IntVar.suggested``.

        Parameters
        ----------
        value : Field
            Suggested step size
        reset : boolean, optional, default : False
            If True, the previous value will be descarded."""
        if reset:
            self._suggested = None
        self.suggested = value if self._suggested is None else np.minimum(
            self._suggested, value)

    @property
    def suggested(self):
        """Suggested step size."""
        if self._suggested is None:
            raise RuntimeError("No step size has been suggested, yet.")
        return self._suggested

    @suggested.setter
    def suggested(self, value):
        if value <= 0:
            raise ValueError(
                "Suggested step size has to be greater than zero.")
        self._suggested = value

    @property
    def stepsize(self):
        '''Current stepsize.'''
        if isinstance(self.updater, Heartbeat):
            return np.minimum(self.updater.beat(self._owner), self.maxstepsize)
        raise RuntimeError(
            "You need to set an Updater for stepsize function first.")

    @property
    def snapshots(self):
        '''Snapshots at which output should be written.

        Even if no outputs are written it needs to contain at least one value that specifies the end point of the
        simulation.'''
        return self._snapshots

    @snapshots.setter
    def snapshots(self, value):
        snaps = np.asarray(value)
        if snaps.size > 1:
            if not all(x < y for x, y in zip(snaps, snaps[1:])):
                raise ValueError("Snapshots have to be strictly increasing")
        self._snapshots = snaps

    @property
    def nextsnapshot(self):
        '''Value of the next snapshot.'''
        if self.snapshots.size < 1:
            raise ValueError("Snapshots are emtpy")
        return self.snapshots[np.argmax(self < self.snapshots)]

    @property
    def prevsnapshot(self):
        '''Value of the previous snapshot.'''
        if self.snapshots.size < 1:
            raise ValueError("Snapshots are emtpy")
        if self < self.snapshots[0]:
            return None
        else:
            return self.snapshots[np.argmin(self >= self.snapshots)-1]

    @property
    def maxstepsize(self):
        '''Maximum possible step size, i.e., to next snapshot.'''
        return self.nextsnapshot - self.getfield(dtype=self.dtype)

    @property
    def prevstepsize(self):
        """Previously taken step size."""
        return self._prevstepsize



import numpy as np

from simframe.frame.field import Field
from simframe.integration.scheme import Scheme
from simframe.utils.color import colorize


class Instruction(Scheme):
    """Integration ``Instruction`` that controls the execution of an integration ``Scheme``."""

    __name__ = "Instruction"

    _Y = None
    _fstep = 1.

    def __init__(self, scheme, Y, fstep=1., controller={}, description=""):
        """Integration instruction

        Parameters
        ----------
        scheme : Scheme
            Integration scheme
        Y : Field
            Variable to be integrated
        fstep : float, optional, default : 1.0
            Fraction of stepsize that this scheme should be used
        controller : dict, optional, default : {}
            Additional keyword arguments passed to integration scheme
        description : str, optional, default : ""
            Description of integration instruction"""
        super().__init__(scheme, controller, description)
        self.Y = Y
        self.fstep = fstep

    @property
    def Y(self):
        '''The ``Field`` to be integrated'''
        return self._Y

    @Y.setter
    def Y(self, value):
        if not isinstance(value, Field):
            raise TypeError("<Y> has to be of type Field.")
        self._Y = value

    @property
    def fstep(self):
        '''Fraction of step size the ``Scheme`` will be applied to ``Field``'''
        return self._fstep

    @fstep.setter
    def fstep(self, value):
        value = np.float(value)
        if value <= 0. or value > 1.:
            msg = "{}: {}".format(
                colorize("Warning", "yellow"), "<fstep> is not in (0, 1].")
            print(msg)
        self._fstep = value

    def __call__(self, dx=None):
        """Execution of the integration instruction

        Parameters
        ----------
        dx : IntVar, optional, default : None
            Stepsize of the integration variable

        Return
        ------
        Y1 : Field
            New value of the variable to be integrated"""
        x0 = self.Y._owner.integrator.var
        Y0 = self.Y
        ret = self.scheme(x0, Y0, self.fstep*dx, **self.controller)
        if ret is False:
            return False
        if ret is True:
            return ret
        else:
            self.Y._buffer = ret
            return True



"""This package contains the core infrastructure of ``simframe``."""

from simframe.frame.abstractgroup import AbstractGroup
from simframe.frame.field import Field
from simframe.frame.frame import Frame
from simframe.frame.group import Group
from simframe.frame.heartbeat import Heartbeat
from simframe.frame.intvar import IntVar
from simframe.frame.updater import Updater

__all__ = ["AbstractGroup",
           "Field",
           "Frame",
           "Group",
           "Heartbeat",
           "IntVar",
           "Updater"]



from simframe.frame import Frame
from simframe.integration import Instruction
from simframe.integration import Integrator
from simframe.integration import schemes
from simframe.io import writers

__name__ = "simframe"
__version__ = "0.4.16"

__all__ = ["Frame",
           "Instruction",
           "Integrator",
           "schemes",
           "writers"]



from setuptools import find_packages
from setuptools import setup
import pathlib

package_name = "simframe"
here = pathlib.Path(__file__).absolute().parent


def read_version():
    with (here / package_name / '__init__.py').open() as fid:
        for line in fid:
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
        else:
            raise RuntimeError("Unable to find version string.")


setup(
    name=package_name,

    description="Framework for scientific simulations",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    keywords="numerical,simulation,integration,science,mathematics,physics",

    url="https://github.com/stammler/simframe",
    project_urls={"Source Code": "https://github.com/stammler/simframe/",
                  "Documentation": "https://simframe.readthedocs.io/"
                  },

    author="Sebastian Stammler, Til Birnstiel",
    author_email="sebastian.stammler@gmail.com, til.birnstiel@lmu.de",
    maintainer="Sebastian Stammler",

    version=read_version(),
    license="GPLv3",

    classifiers=["Development Status :: 3 - Alpha",
                 "Environment :: Console",
                 "Intended Audience :: Developers",
                 "Intended Audience :: Science/Research",
                 "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                 "Natural Language :: English",
                 "Operating System :: OS Independent",
                 "Programming Language :: Python",
                 "Programming Language :: Python :: 3 :: Only",
                 "Topic :: Education",
                 "Topic :: Scientific/Engineering",
                 "Topic :: Scientific/Engineering :: Mathematics",
                 "Topic :: Scientific/Engineering :: Physics",
                 ],

    packages=find_packages(),
    install_requires=["dill", "h5py", "numpy", "pytest", "scipy"],
    include_package_data=True,
    zip_safe=False,
)



class Updater():
    """Class that manages how a ``Group`` or ``Field`` is updated."""

    __name__ = "Updater"

    _func = None

    def __init__(self, func=None):
        """Contains update instructions.

        Parameter
        ---------
        func : callable, optional, default : None
            Function that is called when update function is called. None is a null operation."""
        self._func = func

    def update(self, owner, *args, **kwargs):
        """Function that is called when ``Group`` or ``Field`` to which ``Updater`` belongs is being updated.

        Parameters
        ----------
        owner : Frame
            Parent ``Frame`` object
        args : additional positional arguments
        kwargs : additonal keyword arguments"""
        if self._func is not None:
            return self._func(owner, *args, **kwargs)

    def __str__(self):
        return "{}".format(str(self.__name__))

    def __repr__(self):
        return self.__str__()



from datetime import timedelta
from itertools import cycle
import numpy as np
from time import monotonic
from time import strftime
import sys

from simframe.utils.color import Color


class Progressbar(object):
    """Class for printing progress bar to terminal."""

    _line1 = ""
    _line2 = ""
    _N1 = 0
    _N2 = 0

    def __init__(self, prefix=" ", suffix="| ", fill="▶", empty=" ", length=25, color="blue", spinner=None):
        """This class controls the output of a progress bar on screen.

        Parameters
        ----------
        prefix, string, optional, default : " "
            Initial character of bar
        suffix, string, optional, default : "| "
            Final character of bar
        fill, string, optional, default : "▶"
            Progress character
        empty, string, optional, default : " "
            Character progress has not reached, yet
        length, int, optional, default : 25
            Number of fill/empty characters between prefix and suffix
        color, string or Color, optional, default : "blue"
            Color of bar and spinner
        spinner, Spinner or None, optional, default : None
            Work indicator. If None, standard spinner is used"""
        self._prefix = prefix
        self._suffix = suffix
        self._fill = fill
        self._empty = empty
        self._length = length
        if isinstance(color, Color):
            self._color = color
        else:
            self._color = Color(color)
        if isinstance(spinner, Spinner):
            self._spinner = spinner
        else:
            self._spinner = Spinner()
        # Check if we're in an interactive shell. If not, we don't print progress bar.
        self._print = sys.stdout.isatty()
        # Time keeping for calculating ETA
        self._N_speed = 25
        self._speedbuffer = np.zeros(self._N_speed)
        self._speed = None
        self._t = monotonic()
        self._x = None

    def _getbar(self, filled):
        """Returns actual progress bar between and including prefix and suffix

        Parameters
        ----------
        filled : int
            Number of filled characters in bar.

        Returns
        -------
        bar : string
            String between and including prefix and suffix"""
        bar = "{}{}{}{}".format(
            self._prefix, self._color(filled*self._fill), self._color((self._length-filled)*self._empty), self._suffix)
        return bar

    def _update_speed(self, x):
        '''Updates the current speed.

        Parameters
        ----------
        x : number
            current position of process'''
        t = monotonic()
        if self._x is not None:
            dx = x - self._x
            dt = t - self._t
            speed = dx/dt
            self._speedbuffer = np.insert(self._speedbuffer, 0, speed)[
                :self._N_speed]
            self._speed = self._speedbuffer.mean()
        self._t = t
        self._x = x.copy()

    def _geteta(self, x, s1):
        '''Returns the current ETA

        Parameters
        ----------
        x : number
            current positon of process
        s1 : number
            end position of process

        Returns
        -------
        ETA : string
            string with ETA.'''
        self._update_speed(x)
        if self._speed is None:
            return ""
        eta = (s1-x)/self._speed
        try:
            dt = timedelta(seconds=np.int(eta))
        except:
            dt = "N/A"
        ret = "ETA: {}".format(dt)
        return ret

    def _getline(self, text, x, x0, x1):
        """Returns complete progress bar line

        Parameters
        ----------
        text : string
            Legend of progress bar
        x : number
            Current state of progress
        x0 : number
            Starting point
        x1 : number
            End point

        Returns
        -------
        line : string
            Progress bar line"""
        f = (x-x0)/(x1-x0)
        filled = np.int(np.floor(f*self._length))
        line = "{:10s}{}{:5.1f} %".format(text, self._getbar(filled), f*100.)
        return line

    def _setlines(self, x, x0, x1, s0, s1):
        """Functions sets both progress bar lines for the current snapshot and
        the whole simulation into hidden attribute of class.

        Parameters
        ----------
        x : Number
            Current state of progress
        x0 : Number
            Starting point of snapshot
        x1 : Number
            End point of snapshot
        s0 : Number
            Starting point of simulation
        s1 : Number
            End point of simulation"""
        self._line1 = "{} {}".format(self._getline(
            "Snapshot", x, x0, x1), self._color(self._spinner.next()))
        eta = self._geteta(x, s1)
        self._line2 = "{} {}".format(
            self._getline("Simulation", x, s0, s1), self._color(eta))

    def print(self, x, x0, x1, s0, s1):
        """Function prints the current progress bar. If the end of either the snapshot
        or the simulation is reached, no progress bar will be printed.

        Parameters
        ----------
        x : Number
            Current state of progress
        x0 : Number
            Starting point of snapshot
        x1 : Number
            End point of snapshot
        s0 : Number
            Starting point of simulation
        s1 : Number
            End point of simulation"""
        # Only print if interactive
        if self._print:
            self._N1 = len(self._line1)
            self._N2 = len(self._line2)
            self._setlines(x, x0, x1, s0, s1)
            if not (x == x0 or x == s0):
                self._reset()
            msg = "\x1b[?25l\n{}\n{}".format(self._line1, self._line2)
            print(msg)
            sys.stdout.flush()

    def _reset(self):
        """Resets the current output of progress bar. Function resets the cursor to beginning
        of progress bar and overwrites it with whitespaces, then returns to beginning."""
        # Only print if interactive
        if self._print:
            msg = "\033[3A\r\n{}\n{}\033[3A\r\x1b[?25h".format(
                self._N1*" ", self._N2*" ")
            print(msg)
            sys.stdout.flush()

    def __call__(self, x, x0, x1, s0, s1):
        """Prints the current progress bar.

        Parameters
        ----------
        x : Number
            Current state of progress
        x0 : Number
            Starting point of snapshot
        x1 : Number
            End point of snapshot
        s0 : Number
            Starting point of simulation
        s1 : Number
            End point of simulation"""
        self.print(x, x0, x1, s0, s1)


class Spinner(object):
    """This is a class for displaying some information that the simulation is still runnging."""

    _cycle = None

    def __init__(self, charlist=["●", "○"]):
        """This class is cycling through a list of characters that are displayed after every successfully
        executed timestep.

        Parameters
        ----------
        charlist : list of strings, optional, default : ["●", "○"]
            List of characters to cycle through"""
        self._cycle = cycle(charlist)

    def next(self):
        """Function returns the next character in list"""
        return next(self._cycle)



class Color(object):
    """Class to decorate strings with color tags."""

    _colors = {"black": "\033[30m",
               "red": "\033[91m",
               "green": "\033[92m",
               "yellow": "\033[93m",
               "blue": "\033[94m",
               "purple": "\033[95m",
               "cyan": "\033[96m",
               "white": "\033[98m",
               "reset": "\033[98m"}
    _cstr = "\033[0m"
    _reset = "\033[0m"
    _color = None

    def __init__(self, color="reset"):
        """Class colorizes strings

        Parameters
        ----------
        color : string, optional, default : "reset"
            Default color is user's standard"""
        self.color = color

    def __call__(self, s, color=None):
        """Colorizes string

        Parameters
        ----------
        s : string
            String to be colorized
        color : string, optional, default : None
            Color used to colorize. If None, standard color is used

        Returns
        -------
        cstr : string
            Colorized string"""
        if color is not None and self._checkcolor(color):
            cstr = self._colors[color]
        else:
            cstr = self._cstr
        return "{}{}{}".format(cstr, s, self._reset)

    def _checkcolor(self, color):
        if color in self._colors.keys():
            return True
        else:
            raise ValueError("'{}' is not a valid color.".format(color))

    @property
    def color(self):
        '''Of type ``Color`` with color information.'''
        return self._color

    @color.setter
    def color(self, value):
        if self._checkcolor(value):
            self._color = value
            self._cstr = self._colors[value]


colorize = Color()



import dill


def writedump(object, filename="object.dmp"):
    """Writes object to dump file

    Parameters
    ----------
    object : object
        object to be written to file
    filename : str, optional, default : "object.dmp"
        path to file to be written"""
    with open(filename, "wb") as dumpfile:
        dill.dump(object, dumpfile)


def readdump(filename):
    """Reads dumpfile and returns ``Frame`` object

    Parameters
    ----------
    filename : str
        Path to file to be read

    Returns
    -------
    obj : object
        object read from dump file

    Notes
    -----
    Only read dump files from sources you trust.
    Malware can be injected."""
    with open(filename, "rb") as dumpfile:
        obj = dill.load(dumpfile)
    return obj



from simframe.integration.scheme import Scheme

import numpy as np


def _f_impl_2_midpoint_direct(x0, Y0, dx, jac=None, *args, **kwargs):
    """Implicit 2nd-order midpoint method with direct matrix inversion

    Parameters
    ----------
    x0 : Intvar
        Integration variable at beginning of scheme
    Y0 : Field
        Variable to be integrated at the beginning of scheme
    dx : IntVar
        Stepsize of integration variable
    jac : Field, optional, defaul : None
        Current Jacobian. Will be calculated, if not set
    args : additional positional arguments
    kwargs : additional keyworda arguments

    Returns
    -------
    dY : Field
        Delta of variable to be integrated

    Butcher tableau
    ---------------
     1/2 | 1/2
    -----|-----
         |  1 
    """
    if jac is None:
        jac = Y0.jacobian(x0+0.5*dx)    # Jacobian
    N = jac.shape[0]                    # Problem size
    eye = np.eye(N)                     # Identity matrix

    A = eye - 0.5*dx*jac
    Ainv = np.linalg.inv(A)
    k1 = np.dot(Ainv, np.dot(jac, Y0))

    return dx*k1


impl_2_midpoint_direct = Scheme(
    _f_impl_2_midpoint_direct, description="Implicit 2nd-order direct midpoint method")



from simframe import Frame
from simframe.frame import Field
from simframe.frame import Group
from simframe.frame import IntVar
from simframe.frame import Heartbeat

import pytest


def test_group_description():
    # Note Frame is sub-class of Group
    f = Frame()
    assert f.description == ""
    with pytest.raises(ValueError):
        f.description = None
    f.description = "description"
    assert f.description == "description"
    f = Frame(description="testframe")
    assert f.description == "testframe"


def test_group_updater_func():
    f = Frame()
    assert isinstance(f.updater, Heartbeat)
    f.A = 0
    f.update()
    assert f.A == 0

    def upd(f):
        f.A += 1
    f.updater = upd
    f.update()
    assert f.A == 1


def test_group_updater_list():
    f = Frame()
    f.A = 1
    f.addgroup("g1")
    f.addgroup("g2")
    f.addgroup("g3")

    def g1(f):
        f.A += 1

    def g2(f):
        f.A += 2

    def g3(f):
        f.A *= 3
    f.g1.updater = g1
    f.g2.updater = g2
    f.g3.updater = g3
    ls = ["g1", "g2", "g3"]
    f.updater = ls
    f.update()
    assert f.A == 12
    f.A = 1
    ls = ["g2", "g3", "g1"]
    f.updater = ls
    f.update()
    assert f.A == 10
    f.A = 1
    ls = ["g3", "g2", "g1"]
    f.updater = ls
    f.update()
    assert f.A == 6


def test_group_addgroup():
    f = Frame()
    f.addgroup("g")
    assert isinstance(f.g, Group)


def test_group_addfield():
    f = Frame()
    f.addfield("f", 0.)
    assert isinstance(f.f, Field)
    assert f.f == 0.


def test_group_addintvar():
    f = Frame()
    f.addintegrationvariable("i", 0.)
    assert isinstance(f.i, IntVar)
    assert f.i == 0
